/**************************************************************************
 *
 * Copyright 2002-2006 Columbia University
 * All Rights Reserved
 *
 **************************************************************************/
/* 
 * THINC Extension header - structure definitions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

#ifndef _XTHINCSTR_H_
#define _XTHINCSTR_H_

#include <X11/Xmd.h>

#define XTHINC_EXTENSION_NAME	"THINC"

#define XTHINC_VERSION_MAJOR	0
#define XTHINC_VERSION_MINOR	1

#define xthincNumEvents	0
#define xthincNumErrors	0

/* list of requests */
enum {
	X_THINCQueryVersion = 0,
	X_THINCGetID,		/* 1 */
	X_THINCSetValue,	/* 2 */
	X_THINCGetValue,	/* 3 */
	X_THINCResetValue,	/* 4 */
	X_THINCStatsFlush,	/* 5 */
	X_THINCStatsReset,	/* 6 */
	X_THINCInstrFlush,	/* 7 */
	X_THINCInstrReset,	/* 8 */
	X_THINCWindowShare,	/* 9 */
	X_THINCLast
} X_THINCrequests;

#define XTHINCNumberRequests	X_THINCLast

/* possible values for which (for Get/Set/ResetValue) */
enum {
	xTHINC_val_debugLevel = 0,
	xTHINC_val_mouseDown,
	xTHINC_val_mouseUp,
	xTHINC_val_keybDown,
	xTHINC_val_keybUp,
	xTHINC_val_screensaver,
	xTHINC_val_displayCount,
	xTHINC_val_displayPercent,
	xTHINC_val_nomol	/* end marker */
} xTHINC_vals;

/* a struct to allow for batch gets. easier here than in the protocol :) */
typedef struct {
	int which  B32;	/* get this value */
	int value B32;	/* and the value is this */
} xTHINC_batchGet;

/* the actual requests and replies */

/** Version **/
typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
	CARD16	majorVersion B16;
	CARD16	minorVersion B16;
} xTHINCQueryVersionReq;
#define	sz_xTHINCQueryVersionReq	8

typedef struct {
	CARD8   type;
	CARD8   pad0;
	CARD16  sequenceNumber B16;
	CARD32  length       B32;
	CARD16  majorVersion B16;
	CARD16  minorVersion B16;
	CARD32  pad1         B32;
	CARD32  pad2         B32;
	CARD32  pad3         B32;
	CARD32  pad4         B32;
	CARD32  pad5         B32;
} xTHINCQueryVersionReply;
#define sz_xTHINCQueryVersionReply        32

/** Get ID **/
typedef struct {
	CARD8 reqType;
	CARD8 thincReqType;
	CARD16 length B16;
} xTHINCGetIDReq;
#define sz_xTHINCGetIDReq	4

typedef struct {
	CARD8   type;
	CARD8   pad0;
	CARD16  sequenceNumber B16;
	CARD32  length       B32;
	CARD32  id           B32;	/* your assigned id */
	CARD32  pad1         B32;
	CARD32  pad2         B32;
	CARD32  pad3         B32;
	CARD32  pad4         B32;
	CARD32  pad5         B32;
} xTHINCGetIDReply;
#define sz_xTHINCGetIDReply	32

/** Set Value **/
typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
	CARD32	which B32;	/* set this value */
	CARD32	value B32;	/* to this value */
} xTHINCSetValueReq;
#define	sz_xTHINCSetValueReq	12

typedef struct {
	CARD8   type;
	CARD8   pad0;
	CARD16  sequenceNumber B16;
	CARD32  length       B32;
	CARD32  value        B32;	/* the value to which it got set */
	CARD32  pad1         B32;
	CARD32  pad2         B32;
	CARD32  pad3         B32;
	CARD32  pad4         B32;
	CARD32  pad5         B32;
} xTHINCSetValueReply;
#define sz_xTHINCSetValueReply        32

/** Get Value **/
typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
	CARD32	id B32;		/* slot where to get value from */
	CARD32	which B32;	/* get this value */
} xTHINCGetValueReq;
#define	sz_xTHINCGetValueReq	12

typedef struct {
	CARD8   type;
	CARD8   pad0;
	CARD16  sequenceNumber B16;
	CARD32  length       B32;
	CARD32  which        B32;	/* got this value */
	CARD32  value        B32;	/* and the value is this */
	CARD32  pad1         B32;
	CARD32  pad2         B32;
	CARD32  pad3         B32;
	CARD32  pad4         B32;
} xTHINCGetValueReply;
#define sz_xTHINCGetValueReply        32

/** Reset Value **/
typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
	CARD32	id B32;
	CARD32	which B32;	/* reset this value */
} xTHINCResetValueReq;
#define	sz_xTHINCResetValueReq	12

typedef struct {
	CARD8   type;
	CARD8   pad0;
	CARD16  sequenceNumber B16;
	CARD32  length       B32;
	CARD32  pad1         B32;
	CARD32  pad2         B32;
	CARD32  pad3         B32;
	CARD32  pad4         B32;
	CARD32  pad5         B32;
	CARD32  pad6         B32;
} xTHINCResetValueReply;
#define sz_xTHINCResetValueReply        32

/* flush and reset stats. these are only requests, no replies */
typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
} xTHINCStatsFlushReq;
#define sz_xTHINCStatsFlushReq	4

typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
} xTHINCStatsResetReq;
#define sz_xTHINCStatsResetReq	4

/* flush and reset instrumentation. these are only requests, no replies */
typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
} xTHINCInstrFlushReq;
#define sz_xTHINCInstrFlushReq	4

typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
} xTHINCInstrResetReq;
#define sz_xTHINCInstrResetReq	4

typedef struct {
	CARD8	reqType;
	CARD8	thincReqType;
	CARD16	length B16;
	Window 	window B32;
} xTHINCWindowShareReq;

#define sz_xTHINCWindowShareReq	8

#endif /* _XTHINCSTR_H_ */
