/* Copyright 2007 Ricardo Baratto */
/* Copyright 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/* 
 * THINC Extension -- client side library
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

#include <X11/Xlibint.h>
#include <X11/extensions/Xext.h>
#include <X11/extensions/extutil.h>
#include "X11/extensions/xthincstr.h"

static XExtensionInfo _xthinc_info_data;
static XExtensionInfo *xthinc_info = &_xthinc_info_data;
static char *xthinc_extension_name = XTHINC_EXTENSION_NAME;

#define XTHINCCheckExtension(dpy,i,val) \
	XextCheckExtension(dpy, i, xthinc_extension_name, val)

#define XTHINCGetReq(name,req,info) \
	GetReq (name, req); \
        req->reqType = info->codes->major_opcode; \
        req->thincReqType = X_##name;

#define T_STATUS_FAIL	(Status)0
#define T_STATUS_SUCCESS (Status)1

/*****************************************************************************
 *			   private utility routines                          *
 *****************************************************************************/

#if 0
static int close_display();
static Bool     wire_to_event();
static Status   event_to_wire();
static char *error_string();

static XExtensionHooks xthinc_extension_hooks = {
    NULL,                               /* create_gc */
    NULL,                               /* copy_gc */
    NULL,                               /* flush_gc */
    NULL,                               /* free_gc */
    NULL,                               /* create_font */
    NULL,                               /* free_font */
    close_display,                      /* close_display */
    wire_to_event,                      /* wire_to_event */
    event_to_wire,                      /* event_to_wire */
    NULL,                               /* error */
    error_string                        /* error_string */
};
#else
static XExtensionHooks xthinc_extension_hooks = {
    NULL,                               /* create_gc */
    NULL,                               /* copy_gc */
    NULL,                               /* flush_gc */
    NULL,                               /* free_gc */
    NULL,                               /* create_font */
    NULL,                               /* free_font */
    NULL,                               /* close_display */
    NULL,                               /* wire_to_event */
    NULL,                               /* event_to_wire */
    NULL,                               /* error */
    NULL                                /* error_string */
};
#endif

static XEXT_GENERATE_FIND_DISPLAY (find_display, xthinc_info, 
	xthinc_extension_name, &xthinc_extension_hooks, xthincNumEvents, 
	NULL)

#if 0
static XEXT_GENERATE_CLOSE_DISPLAY (close_display, xthinc_info)

static char *xthinc_error_list[] = {
	""
};

static XEXT_GENERATE_ERROR_STRING (error_string, xthinc_extension_name,
				   xthincNumErrors, xthinc_error_list)
#endif

/*****************************************************************************
 *		    public THINC Extension routines                          *
 *****************************************************************************/
Bool XTHINCQueryExtension(Display *dpy)
{
	XExtDisplayInfo *info = find_display(dpy);

	if(XextHasExtension(info)) {
		return True;
	} else {
		return False;
	}
}

Bool XTHINCQueryVersion(Display *dpy, int *major_version_return, 
	int *minor_version_return)
{
	XExtDisplayInfo *info = find_display (dpy);
	xTHINCQueryVersionReply rep;
	register xTHINCQueryVersionReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCQueryVersion refers to the #define's in xthincstr.h */
	XTHINCGetReq(THINCQueryVersion, req, info);
	if (!_XReply(dpy, (xReply *)&rep, 0, xFalse)) {
		UnlockDisplay(dpy);
		SyncHandle();
		return False;
	}
	*major_version_return = rep.majorVersion;
	*minor_version_return = rep.minorVersion;
	UnlockDisplay(dpy);
	SyncHandle();

	if(*major_version_return != XTHINC_VERSION_MAJOR)
		return False;
	else 
		return True;
}

Status XTHINCGetID(Display *dpy, int *id_return)
{
	XExtDisplayInfo *info = find_display (dpy);
	xTHINCGetIDReply rep;
	register xTHINCGetIDReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCGetID refers to the #define's in xthincstr.h */
	XTHINCGetReq(THINCGetID, req, info);
	if (!_XReply(dpy, (xReply *)&rep, 0, xFalse)) {
		UnlockDisplay(dpy);
		SyncHandle();
		return T_STATUS_FAIL;
	}
	*id_return = rep.id;

	UnlockDisplay(dpy);
	SyncHandle();

	return T_STATUS_SUCCESS;
}

/* use it to change an integer setting in the server. see xthincstr.h for
 * the possible settings */
Status XTHINCSetValue(Display *dpy, int which, int value)
{
	XExtDisplayInfo *info = find_display (dpy);
	xTHINCSetValueReply rep;
	register xTHINCSetValueReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCSetValue refers to the #define's in xthincstr.h */
	XTHINCGetReq(THINCSetValue, req, info);
	req->which = which;
	req->value = value;
	if (!_XReply(dpy, (xReply *)&rep, 0, xFalse)) {
		UnlockDisplay(dpy);
		SyncHandle();
		return T_STATUS_FAIL;
	}

	UnlockDisplay(dpy);
	SyncHandle();

	return T_STATUS_SUCCESS;
}

/* use it to get the value of an integer setting in the server. see 
 * xthincstr.h for the possible settings */
Status XTHINCGetValue(Display *dpy, int id, int which, int *value_return)
{
	XExtDisplayInfo *info = find_display (dpy);
	xTHINCGetValueReply rep;
	register xTHINCGetValueReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCGetValue refers to the #define's in xthincstr.h */
	XTHINCGetReq(THINCGetValue, req, info);
	req->id = id;
	req->which = which;
	if (!_XReply(dpy, (xReply *)&rep, 0, xFalse)) {
		UnlockDisplay(dpy);
		SyncHandle();
		return T_STATUS_FAIL;
	}
	*value_return = rep.value;

	UnlockDisplay(dpy);
	SyncHandle();

	return T_STATUS_SUCCESS;
}

/* use it to reset the value of an integer setting in the server. see 
 * xthincstr.h for the possible settings */
Status XTHINCResetValue(Display *dpy, int id, int which)
{
	XExtDisplayInfo *info = find_display (dpy);
	xTHINCResetValueReply rep;
	register xTHINCResetValueReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCResetValue refers to the #define's in xthincstr.h */
	XTHINCGetReq(THINCResetValue, req, info);
	req->id = id;
	req->which = which;
	if (!_XReply(dpy, (xReply *)&rep, 0, xFalse)) {
		UnlockDisplay(dpy);
		SyncHandle();
		return T_STATUS_FAIL;
	}

	UnlockDisplay(dpy);
	SyncHandle();

	return T_STATUS_SUCCESS;
}

/* for now we do the loop on the client side. maybe in the future we'll add
 * another protocol message and move this functionality to the server. or not */
Status XTHINCGetValueBatch(Display *dpy, int id, xTHINC_batchGet* vals, 
	int count)
{
	register int i;
	register xTHINC_batchGet *c;

	c = vals;
	for(i=0; i<count; ++i, ++c) {
		if(XTHINCGetValue(dpy, id, c->which, &c->value) == T_STATUS_FAIL)
		{
			return T_STATUS_FAIL;
		}
	}
	return T_STATUS_SUCCESS;
}

/* for now we do the loop on the client side. maybe in the future we'll add
 * another protocol message and move this functionality to the server. or not */
Status XTHINCResetValueBatch(Display *dpy, int id, int* vals, int count)
{
	register int i;
	register int *c;

	c = vals;
	for(i=0; i<count; ++i, ++c) {
		if(XTHINCResetValue(dpy, id, *c) == T_STATUS_FAIL) {
			return T_STATUS_FAIL;
		}
	}
	return T_STATUS_SUCCESS;
}

Bool XTHINCStatsFlush(Display *dpy)
{
	XExtDisplayInfo *info = find_display (dpy);
	register xTHINCStatsFlushReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCStatsFlush refers to the #define's (X_blah) in xthincstr.h */
	XTHINCGetReq(THINCStatsFlush, req, info);
	UnlockDisplay(dpy);
	SyncHandle();

	return True;
}

Bool XTHINCStatsReset(Display *dpy)
{
	XExtDisplayInfo *info = find_display (dpy);
	register xTHINCStatsResetReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCStatsReset refers to the #define's (X_blah) in xthincstr.h */
	XTHINCGetReq(THINCStatsReset, req, info);
	UnlockDisplay(dpy);
	SyncHandle();

	return True;
}

Bool XTHINCInstrFlush(Display *dpy)
{
	XExtDisplayInfo *info = find_display (dpy);
	register xTHINCInstrFlushReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCInstrFlush refers to the #define's (X_blah) in xthincstr.h */
	XTHINCGetReq(THINCInstrFlush, req, info);
	UnlockDisplay(dpy);
	SyncHandle();

	return True;
}

Bool XTHINCInstrReset(Display *dpy)
{
	XExtDisplayInfo *info = find_display (dpy);
	register xTHINCInstrResetReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCInstrReset refers to the #define's (X_blah) in xthincstr.h */
	XTHINCGetReq(THINCInstrReset, req, info);
	UnlockDisplay(dpy);
	SyncHandle();

	return True;
}

/* Tell the THINC server to share window w (and it's children) */
Bool XTHINCWindowShare(Display *dpy, Window w)
{
	XExtDisplayInfo *info = find_display (dpy);
	register xTHINCWindowShareReq *req;

	XTHINCCheckExtension (dpy, info, False);

	LockDisplay(dpy);
	/* THINCInstrReset refers to the #define's (X_blah) in xthincstr.h */
	XTHINCGetReq(THINCWindowShare, req, info);
	req->window = w;
	UnlockDisplay(dpy);
	SyncHandle();

	return True;
}
