/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/*
 * Header for thinc module (Remote display ala VNC - SunRay)
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

#ifndef _THINC_H_
#define _THINC_H_

/* #defines to control various additional (experimental) features of THINC */

/* Allow remote display of just particular windows */
#if 0
#define T_APPSHARE
#endif

/* do recording of commands */
#if 1
#define T_RECORD
#endif

/* don't do some drawing on the server (not ready for prime time use) */
#if 0
#define T_NODRAW
#endif

/* define to disable unnecessary things that may slow down the server. good
 * for benchmarking or production use */
#if 1
#define T_WICKED_FAST
#endif

/* Do additional sanity checks */
/* TODO: Should probably have these on all the time... */
#ifndef T_WICKED_FAST
#define T_STRICT_CHECKS
#endif

/* Additional runtime debugging/testing config options */
#ifndef T_WICKED_FAST
#define T_ENABLE_DEBUG_OPTIONS
#endif

/* Define to have the server collect statistics */
#ifndef T_WICKED_FAST
#define T_GETSTATS
#endif

/* comment to disable debugging altogether */
#ifndef T_WICKED_FAST
#define T_DO_DEBUG
#endif

/* enable instrumentation code (different from stats, which is why is not
 * protected by T_WICKED_FAST!) */
#if 0
#define T_INSTRUMENT
#endif

/* compression stuff */
/* comment to use zlib instead of png */
#define T_USEPNG

/* MIT-SHM */
/* comment this if you don't want THINC to support the MIT Shared Memory
 * extension. Note that this may make THINC not work properly when 
 * Offscreen Pixmaps are being used and applications make use of this
 * extension. */ 
#define T_USEMITSHM

/* Define to support XVideo */
#if 0
#ifdef XvExtension
#define T_USEXVIDEO
#endif
#else
/* XXX: New build system doesn't define XvExtension (or i don't know how
 * to get it) */
#define T_USEXVIDEO
#endif

/* Implement XAA acceleration. XXX Alpha code, don't use XXX */
/*
#define T_USEXAA
 */

/* Support RENDER (see thincRender.c) */
#ifdef RENDER
#define T_USERENDER
#endif

/* end #defines */

/* we need this for wrapper functions */
#include "scrnintstr.h"
#include "gcstruct.h"
#include "windowstr.h"
#include "regionstr.h"
#include "dixfontstr.h"
#include "xorg/xf86Cursor.h"

/* our headers */
#include "thincCommon.h"
#include "thincProto2.h"
#include "thincUtil.h"
#include "thincGC.h"
#include "thincInstrument.h"

#ifdef T_USERENDER
#include "picturestr.h"
#include "mipict.h"
#endif

#ifdef T_USEXAA
#include "xaa.h"
#include "thincAccel.h"
#endif

/* security */
#include "thinc_ssl.h"

/* cache */
#include "thinc_cache.h"
#define T_RAWCACHE_BITS	10
#define T_BITCACHE_BITS	16
#define T_PIXCACHE_BITS	4

/* resize */
#include "thinc_resize.h"

/* linked list */
#include "thincList.h"

/* Version #defines */
#define THINC_NAME		"THINC"
#define THINC_DRIVER_NAME	"thinc"
#define THINC_VERSION_NAME	"0.1"
#define THINC_VERSION_MAJOR	0
#define THINC_VERSION_MINOR	1
#define THINC_PATCHLEVEL	2
#define THINC_VERSION_CURRENT	((THINC_VERSION_MAJOR << 24) | \
     				(THINC_VERSION_MINOR << 16) | \
				THINC_PATCHLEVEL)

/* Command list size */
#define T_CMDINFOLIST_SIZE	20

/* Default Compression Level. */
#define T_ZCOMPRESS_LEVEL	Z_BEST_SPEED

/* forward declarations */
struct thincClient_S;
struct thincCmdInfo_S;
struct thincCmdList_S;
struct thincVideo_S;
struct thincEncode_S;
struct thincCmd_S;
#ifdef T_APPSHARE
struct thincApp_S;
#endif
struct thincCursor_S;
#ifdef T_RECORD
struct thincRecord_S;
#endif

#if 1
struct thincScreenStats {
	struct thincList_S	list;
	int			id;
	int			mouse_down;
	int			mouse_up;
	int			keyb_down;
	int			keyb_up;
	int			display_count;
	RegionRec		display_change;
};
#endif

/* Screen private structure. One of this for each screen in the system */
typedef struct thincScreen_S {
	/* The Screen to which we belong */
	ScreenPtr		pScreen;

	/* client list */ 
	struct thincList_S 	clients;
#ifdef T_NODRAW
	struct thincClient_S*	client0;	/* the client that's not
						   really a client */
#endif	/* T_NODRAW */

	/* network */
	int 			listenSocket;
	fd_set 			fds;	/* select list */
	int 			fdmax;

	/* framebuffer info */
	/* TODO: store the width of the framebuffer in bytes */
	char *		theFB;	/* the frame buffer */
	unsigned int		width:16;
	unsigned int		height:16;
	unsigned int		widthBytes:32;

	unsigned int		depth:8;  /* color depth */
	unsigned int		bpp:8;	/* bits per pixel */
	unsigned int		Bpp:8;	/* Bytes per pixel */
	unsigned int		Bpp_shift:8;	/* hack */

	unsigned int		updateWidth;
	unsigned int		updateHeight;
	unsigned int		updateSize;

	/* debugging level for this screen. Specified in config file */
	unsigned int		debugLevel:16;
	/* number of clients fully connected */
	unsigned int		clientCount:16;

	/* Message Compression */
	struct thincEncode_S*		encoder;

	/* options. settable in config file */
	/* IMPORTANT: Note padding below and change appropriately as you add 
	 * or remove options */
	unsigned int		beSecure:1;
	unsigned int		useAuth:1;
	unsigned int		useVideo:1;
	unsigned int		useOffscreen:1;
	unsigned int		useCopy:1;
	unsigned int		useSFill:1;
	unsigned int		usePFill:1;
	unsigned int		useGlyph:1;
	unsigned int		useBilevel:1;

	unsigned int		useHWCursor:1;
	unsigned int		doWriteBuffer:1;
	unsigned int		doCmdBuffer:1;
	unsigned int		compress:1;
	unsigned int		compressVideo:1;
	unsigned int		allowResize:1;
	unsigned int		doSched:1;

	unsigned int		useRawCache:1;
	unsigned int		useBitCache:1;
	unsigned int		usePixCache:1;

	unsigned int		doRecord:1;
	
	/* ==== debug,testing options ==== */
#ifdef T_ENABLE_DEBUG_OPTIONS
	/* This one tells the server that if an empty command queue is
	 * found during a CopyArea(), it should inject/issue a RAW Update
	 * to account for the copy. Good and Bad: Good -> more tolerant to
	 * unknown drawing commands. Bad -> We'll never get to figure out
	 * those drawing commands (probably missing potential opportunities
	 * for compression) and thinc will be buggy forever. Note that
	 * this only makes sense if Offscreen is activated */
	unsigned int		emptyQueuesOK:1;
	/* XXX For testing */
	unsigned int		doRaw:1;
	unsigned int		allowOverlap:1;
	/********** Update these two pads as needed ************/
	unsigned int		unused:9;
#else	/* DEBUG_OPTIONS */
	unsigned int		unused:12;
#endif

	/* array of thincCmd structs, which documents the function used
	 * to handle a particular protocol command (see thincPixmap.h for
	 * details). Indexed with (command_id - 10) (first fb command's 
	 * id -> 11) */
	struct thincCmdInfo_S*		cmdHandlerList[T_CMDINFOLIST_SIZE];

	/** XVideo **/
	struct thincVideo_S*		thincVid;
	unsigned long 			droppedFrames;

	/* Timer for buffer */
	OsTimerPtr			timer;

	/* Security stuff */
	SSL_CTX				*ctx;

	/* command recording stuff */
#ifdef T_RECORD
	struct thincRecord_S		*record;
	unsigned int			rate_ms;
#endif

	/* Cache */
	unsigned int			raw_cache_bits;
	unsigned int			bit_cache_bits;
	unsigned int			pix_cache_bits;

	/* Ping */
	unsigned int			ping_freq;

#ifdef T_GETSTATS
	/***** stats *****/
	/* per command */
	struct thincStats		rawStats;
	struct thincStats		copyStats;
	struct thincStats		sfillStats;
	struct thincStats		pfillStats;
	struct thincStats		glyphStats;
	struct thincStats		bilevelStats;

	/* draw stats */
	unsigned long		countFillSpans;
	unsigned long		countSetSpans;
	unsigned long		countPutImage;
	unsigned long		countCopyArea;
	unsigned long		countCopyPlane;
	unsigned long		countPolyPoint;
	unsigned long		countPolylines;
	unsigned long		countPolySegment;
	unsigned long		countPolyRectangle;
	unsigned long		countPolyArc;
	unsigned long		countFillPolygon;
	unsigned long		countPolyFillRect;
	unsigned long		countPolyFillArc;
	unsigned long		countPolyText8;
	unsigned long		countPolyText16;
	unsigned long		countImageText8;
	unsigned long		countImageText16;
	unsigned long		countImageGlyphBlt;
	unsigned long		countPolyGlyphBlt;
	unsigned long		countPushPixels;
	unsigned long 		countPaintWindowBorder;
	unsigned long		countPaintWindowBackground;
	unsigned long		countCopyWindow;

	/* how many times we merged commands */
	unsigned long		merge_try;
	unsigned long		merges;
	unsigned long		merge_raw;
	unsigned long 		merge_sfill;
	unsigned long		merge_time;
	unsigned long		merge_time_alloc;
	unsigned long		merge_time_free;
	/* - overwrites is how many times we overwrote a command.
	 * - try is how many times we tried to find overlap 
	 * - any is how many times we called overwrite and it actually
	 * overwrote something
	 * - complete is how many times a command was fully overwritten
	 */
	unsigned long		overwrite_try;
	unsigned long		overwrite_try_cmds;
	unsigned long		overwrite_any;
	unsigned long		overwrite_any_cmds;
	unsigned long		overwrites;
	unsigned long		overwrite_complete;
	unsigned long		overwrite_noover;
	unsigned long		overwrite_trans;

	unsigned long		overwrite_time;

	/* how many times we tried to find dependencies */
	unsigned long		find_deps;

	/* how many of each type of command we overwrote */
	/* XXX: Hackishly reusing T_CMDINFOLIST_SIZE */
	unsigned long		overwrite_type[T_CMDINFOLIST_SIZE];
#endif	/* T_GETSTATS */

	/*** Wrapped Functions ***/
	ScreenWakeupHandlerProcPtr 	WakeupHandler;
	ScreenBlockHandlerProcPtr	BlockHandler;
	CloseScreenProcPtr 		CloseScreen;

#ifdef T_NODRAW
	SaveScreenProcPtr		SaveScreen;
#endif

	/* for multiple servers */
	xf86EnableDisableFBAccessProc   *EnableDisableFBAccess;

	/* window funcs */
#ifdef T_APPSHARE	/* no need to wrap this normally */
	CreateWindowProcPtr 		CreateWindow;
	DestroyWindowProcPtr 		DestroyWindow;
	PositionWindowProcPtr 		PositionWindow;
	RealizeWindowProcPtr 		RealizeWindow;
	UnrealizeWindowProcPtr 		UnrealizeWindow;
	ChangeWindowAttributesProcPtr 	ChangeWindowAttributes;
	WindowExposuresProcPtr 		WindowExposures;
#endif
	PaintWindowBackgroundProcPtr 	PaintWindowBackground;
	PaintWindowBorderProcPtr 	PaintWindowBorder;
	CopyWindowProcPtr 		CopyWindow;
	/* pixmap funcs */
	CreatePixmapProcPtr		CreatePixmap;
	DestroyPixmapProcPtr		DestroyPixmap;
	ModifyPixmapHeaderProcPtr	ModifyPixmapHeader;
	/* gc funcs */
	CreateGCProcPtr			CreateGC;

#ifdef T_USERENDER
	/* Render Procs we need to wrap */
	CreatePictureProcPtr		CreatePicture;
	DestroyPictureProcPtr		DestroyPicture;
	ChangePictureClipProcPtr	ChangePictureClip;
	DestroyPictureClipProcPtr	DestroyPictureClip;
	
	ChangePictureProcPtr		ChangePicture;
	ValidatePictureProcPtr		ValidatePicture;
	
	InitIndexedProcPtr		InitIndexed;
	CloseIndexedProcPtr		CloseIndexed;
	UpdateIndexedProcPtr		UpdateIndexed;

	/* the rendering functions */
	CompositeProcPtr		Composite;
	GlyphsProcPtr			Glyphs;
	CompositeRectsProcPtr		CompositeRects;
	TrapezoidsProcPtr		Trapezoids;
	TrianglesProcPtr		Triangles;
	TriStripProcPtr			TriStrip;
	TriFanProcPtr			TriFan;
	RasterizeTrapezoidProcPtr	RasterizeTrapezoid;

#endif /* T_USERENDER */

#ifdef T_USEXAA
	/* XAA */
	XAAInfoRecPtr			xaaInfo;
	Bool				xaaPictureWrite;
	thinc_rectRegion		xaaPictureInfo;

	char*			xaaAperture;
	int				xaaApertureSize;
#endif /* T_USEXAA */

	/* Bitmap function pointers (depend on color depth) 
	 * (see thincBits.c) */
	int (*bitmap_from_fb)(struct thincScreen_S * , char *, 
		char *, xRectangle *, int , unsigned long );
	int (*bitmap_bilevel)(struct thincScreen_S *, char *, 
		char *, xRectangle *, int, unsigned long, 
		unsigned long *);

	/* HW Cursor */
	struct thincCursor_S*	cursor;
	/* the cursor may be moved for 2 reasons:
	 * 1) user moving the mouse on the client side
	 * 2) an application making the mouse jump to a specified position
	 *
	 * If SetCursorPosition is called in response to 1) there's no
	 * need to make the client move the cursor and it's actually
	 * undesirable to do so. However, if it's in response to 2) we
	 * need to inform the client of this. The idea behind moveCount
	 * is that every time a move mouse event comes in moveCount is 
	 * increased, and when SetCursorPosition is called, moveCount
	 * is decreased. If on a SetCursorPosition, moveCount is 0,
	 * then there are no outstanding movement events and the call
	 * is caused by 2), so we send a message to the client.
	 * XXX: An application could insert a jump in between mouse moves.
	 * however, how good it'd be to make the mouse jump to a 
	 * different position in the middle of a mouse move?
	 * Must play and see...
	 */
	unsigned int		moveCount:16;

	/* how many clients are resized */
	unsigned int		resizeCount:16;

	struct thincClient_S *	last_mouse_client;

	/* owner of the session */
	unsigned int 		owner_len;
	char *			owner;

#ifdef T_APPSHARE
	struct thincApp_S*	apps;	/* thincAppshare.h */
	unsigned int		app_count;	/* how many apps are active */
#endif
	
	/* some statistics for thincback's checkpoint control */
#if 0
	int			mouse_down;
	int			mouse_up;
	int			keyb_down;
	int			keyb_up;
	int			display_count;
	RegionRec		display_change;
#else
	int			screensaver;	/* global */
	double			total_area;

	/* list of thincScreenStats */
	struct thincList_S	stats;
	int			stats_next_id;
#endif
} thincScreenRec, *thincScreenPtr;

/* holds all information on the cursor */
typedef struct thincCursor_S {
	/* FIXME: if we build static server, this is not needed (i think) */
	xf86CursorInfoPtr 	info;	/* pointer to xf86 cursor
						   info struct */
	
	unsigned int		width :8;
	unsigned int		height:8;
	unsigned int		size  :16; 	/* the cursor size (bytes) */
	unsigned int		x :16;		/* current cursor position */
	unsigned int		y :16;
	unsigned short		xhot;		/* hot spot location */
	unsigned short		yhot;
	unsigned long		fg;		/* cursor colors */
	unsigned long		bg;
	char *		cursor;		/* the cursor. source, mask*/
	/* flags */
	unsigned int		on:1;		/* is cursor visible? */
	unsigned int		argb:1;		/* is the current cursor argb */
	unsigned int		pad:14;
} thincCursorRec, *thincCursorPtr;

/* GC Private */
typedef struct thincGC_S {
	/* Use to pass information from one function to the other
	 * using the same GC. Current flags defined below */
	int			flags;

	/* The GCOps from the driver. Our GC functions wrap around this */
	GCOps		*driver_ops;

	/* Wrapped ValidateGC function */
	GCFuncs		*driver_funcs;
} thincGCRec, *thincGCPtr;

/* Flags used for thincGCPtr->flags */
#define gc_none			0x0000

#define gc_clearflag(pThincGC, flag) \
	pThincGC->flags &= ~(flag)

/* Some drawing operations have high level and low level components
 * in the GC ops vector (e.g. the text drawing functions). This field 
 * is supposed to be used as a communication mechanism by those 
 * functions, to, for example, stop the high level from sending an 
 * update, when the lowlevel function already sent it */
#define gc_lowlevelUpdate	0x0001

/* Used by ValidateGC to tell the drawing functions that the GC has
 * weird drawing settings and they should just bail out on trying to
 * optimize */
#define gc_weirdo		0x0002

/* devPrivates indexes */
extern int thincPrivateIndex;	/* ScreenPtr index */
extern int thincPixmapPrivateIndex;	/* PixmapPtr index */
extern int thincGCPrivateIndex;	/* GCPtr index */
#ifdef T_APPSHARE
extern int thincWindowPrivateIndex;	/* WindowPtr index */
#endif

/* Macros to extract private structures */
#define THINCPtr(pScreen) \
	(thincScreenPtr)((pScreen)->devPrivates[thincPrivateIndex].ptr)
#define THINCGCPtr(pGC) \
	(thincGCPtr)((pGC)->devPrivates[thincGCPrivateIndex].ptr)
#define THINCPIXPtr(pPixmap) \
	(thincPixmapPtr)((pPixmap)->devPrivates[thincPixmapPrivateIndex].ptr)
#ifdef T_APPSHARE
#define THINCWINPtr(pWindow) \
	(thincWindowPtr)((pWindow)->devPrivates[thincWindowPrivateIndex].ptr)
#endif

/****************************** Client ***********************/

/* msg buffer's size is based on a full TCP packet bounded by an Ethernet 
 * Frame */
#define T_ETH_MTU	1514
#define	T_ETH_HEADER	14
#define T_ETH_MAX	(T_ETH_MTU - T_ETH_HEADER)

#define T_IP_HEADER	20
#define T_IP_MAX	(T_ETH_MAX - T_IP_HEADER)

#define T_TCP_HDR	20
#define T_TCP_MAX	(T_IP_MAX - T_TCP_HDR)

/* max size for msg buffer (write buffer) */
#define T_MAXSIZE	T_TCP_MAX

#define T_STATIC_BUFLEN	T_MAXSIZE
/* FIXME: For non-blocking to work, buffer size cannot be big. */
#if 1
#define T_WRITE_BUFLEN	T_MAXSIZE*5
#else
#define T_WRITE_BUFLEN	T_MAXSIZE*1000
#endif
/* size of read buffer. Increase as needed */
#define T_READ_BUFLEN	256

/* computes max size of data part for a message of type msg than can fit
 * in an empty buffer. The header and msg size (from thincProto.h) are 
 * added automatically. (msg should be something like T_FB_RUPDATE -> 
 * see thincProto.h) */
/* XXX: DEPRECATED. Use T_CL_TOTALDATA/T_CL_TOTALMSGDATA */
#define T_MAXDATA(msg)	\
	(T_WRITE_BUFLEN - THINC_HEADERSIZE - msg##_SIZE)

#define T_CL_TOTALDATA(cl) \
	((cl)->msgBufMaxSize)

#define T_CL_TOTALMSGDATA(cl, msg) \
	((cl)->msgBufMaxSize - THINC_HEADERSIZE - msg##_SIZE)

/* returns max size of data part for a message of type msg that can
 * fit in the current client buffer. Header size is added automatically.
 * If msg is not 0, the msg size (as found in thincProto.h) is also added 
 * in (msg should be something like T_FB_RUPDATE -> see thincProto.h) */
#define T_CL_MAXMSGDATA(cl, msg) \
	((cl)->msgBufMaxSize - (cl)->msgBufSize - THINC_HEADERSIZE - \
	 msg##_SIZE)

#define T_CL_MAXDATA(cl) \
	((cl)->msgBufMaxSize - (cl)->msgBufSize )

/* if less than this space in buffer, functions may flush buffer and query
 * for space again */
#define T_CL_MINSIZE	20

#ifdef T_USEXVIDEO
struct thincClVideo_S;
#endif

struct thincSched_S;
struct thincMsgStack_S;
struct thincMsgStackStep_S;

#define T_CLID_LEN	32

typedef struct thincClient_S {
	struct thincList_S	link;
	thincScreenPtr 		parent;

	/** I/O **/
	int 			sock;
	SSL			*ssl;
	int	(*read)  (struct thincClient_S*, void *, 
			unsigned int);
	int	(*write) (struct thincClient_S*, const void *, 
			unsigned int);

	int	(*buffer)(struct thincClient_S*, struct thincCmd_S* );
	int	(*flush) (thincScreenPtr, struct thincClient_S*);
	int	(*resize)(char *, char *, 
			unsigned int, unsigned int,	/* width */
			unsigned int, unsigned int,	/* height */
			unsigned int, unsigned int);	/* stride */
	int	(*send_msg)(struct thincClient_S*, 
			int /* type */,
			unsigned int /* msg_len */,
			int /* flags */
		       );

	void	(*close_cl)(struct thincClient_S*);

	/* normally the ip address */
	char			id[T_CLID_LEN];

	/* message stacks */
	struct thincMsgStack_S*	rawStack;
	struct thincMsgStack_S* bitmapStack;
	struct thincMsgStack_S* pixmapStack;
	struct thincMsgStack_S* fillStack;
	struct thincMsgStack_S* copyStack;

	/* TODO: Need message stack for handshake */

	/* Write buffering. Two layers:
	 * => Buffer
	 * => Flush
	 *
	 * The Buffer layer buffers actual commands. Useful for slow
	 * clients/networks
	 * The Flush layer buffers bytes. Useful for good packet
	 * utilization.
	 * See thincBuffer.h for more details
	 */
	/* == flush layer ==
	 * Need to do it here (and not globally in thincScreenPtr) because 
	 * each client may have different messages (for example, when they 
	 * have different clipping regions). */
	char		msgBuffer[T_WRITE_BUFLEN];
	char		*msgBufP; 	/* current tail of the buffer 
					   (or where to write data) */
	unsigned int	msgBufMaxSize;	/* max size of the buffer */
	unsigned int	msgBufSize;	/* actual size of the buffer 
					   (or how much data is there) */
	/* == buffer layer ==
	 * Command buffer */ /* see also flags below */
	struct thincList_S	cmdBuffer;

	/* cache */
	thinc_cache*			raw_cache;
	thinc_cache*			bit_cache;
	thinc_cache*			pix_cache;

	/* per-client read buffer. 
	 * Note that this buffer is used for optimization (to cut on the 
	 * number of read() calls made), not to read any single message 
	 * that's sent by a client. */
	char			rBuffer[T_READ_BUFLEN];
	char *			rBufStart;
	unsigned int		rBufSize;

	/* clipping region */
	RegionRec		clClip;

	/* command scheduler */
	struct thincSched_S	*sched;
	unsigned long		last;

	/*** Video ***/
#ifdef T_USEXVIDEO
	struct thincClVideo_S		*video;
#endif

	/*** Resize support ***/
	unsigned int	curr_width;
	unsigned int	curr_height;
	double		factor_x;
	double		factor_y;

	/* flags, 16 bits total */
	unsigned int		isSecure:1;	/* is this client using SSL? */
	unsigned int		clipIsSub:1; 	/* true if the clipping region
						 * is not equal to the size of 
						 * the framebuffer */
	unsigned int		resized:2;	/* are we resized? */
	unsigned int		buffer_batch:1;
	unsigned int		print_msg:1;	/* print message when an
						   error occurs */
	unsigned int		doVideo:1;
	unsigned int		doAuth:1;
	unsigned int		connected:1;
	unsigned int		pad:7;

	struct thincMsgStackStep_S* connectStack;
	unsigned int		sent;
#ifdef T_RECORD
	unsigned long		last_flush;
#endif
} thincClientRec, *thincClientPtr;

/*************************** functions ************************************/

/* thinc.c */
Bool ThincInit(ScreenPtr pScreen, char *FBStart);
int ThincCursorInit(ScreenPtr pScreen);
void thincWakeupHandler	(int i, pointer blockData, unsigned long err,
			 pointer pReadmask);
void thincBlockHandler(int i, pointer blockData, pointer pTimeout,
	pointer pReadmask);

Bool thincCloseScreen	(int index, ScreenPtr pScreen);

/* thincClient.c */
int thincNewConnection	(thincScreenPtr pThinc);
void thincAddSocket	(thincScreenPtr pThinc, int sock);
void thincRemoveSocket	(thincScreenPtr pThinc, int sock); 
int thincHandleClient	(thincClientPtr cl);
void thincCloseClient	(thincClientPtr cl);
int thincHaveNewClients(thincScreenPtr pThinc, fd_set *socks);
int thincCheckClients(thincScreenPtr pThinc, fd_set *socks, int num);
#ifdef T_NODRAW
int thincClient0Init(thincScreenPtr pThinc);
void thincClient0Close(thincClientPtr cl);
int thincClient0Purge(thincScreenPtr pThinc);
#endif
int thinc_v2SendReply(thincClientPtr cl, int reply_size);

/* helper. creates a region and calls thincSendFBUpdateAllRegion */
void thincSendFBUpdateAll(thincScreenPtr pThinc, int x, int y, int width, 
	int height);

/* thincBits.c */
int thincInitBitmaps(thincScreenPtr pThinc);
int thinc_createBitmapFB(thincScreenPtr pThinc, char *bitmap, 
	xRectangle *area, unsigned long fg);
int thinc_createBitmapPixmap(thincScreenPtr pThinc, PixmapPtr pix,
	char *bitmap, xRectangle *area, unsigned long fg);
unsigned int thinc_getBitmapRect(thincScreenPtr pThinc, PixmapPtr pixmap,
	xRectangle *area, register unsigned long fg, xRectangle *out);
int thinc_createBitmapFBBilevel(thincScreenPtr pThinc, char *bitmap, 
	xRectangle *area, unsigned long fg, unsigned long *bg);
int thinc_createBitmapPixmapBilevel(thincScreenPtr pThinc, PixmapPtr pix,
	char *bitmap, xRectangle *area, unsigned long fg, 
	unsigned long *bg);

/* thincMouse.c */
void thincInitMouse	(void);

/* thincKeyb.c */
void thincInitKeyboard	(void);
int thincHandleMotionEv	(thincClientPtr cl);
int thincHandleButtonEv	(thincClientPtr cl);
int thincHandleKeybEv	(thincClientPtr cl);

/* thincUtil.c */
int thincReadHeader(thincClientPtr cl, thinc_header *h);
int thincReadAnyMsg(thincClientPtr cl, void *msg, unsigned int msg_size);
int thincReadData(thincClientPtr cl, void *buf, unsigned int count);
int thincReadDataNoblock(thincClientPtr cl, void *data, unsigned int count);
int thincBadRead(thincClientPtr cl, void *buf, unsigned int count);
int thincSendMsg(thincClientPtr cl, int type, unsigned int msg_len, int flags);
int thincSendMsgNull(thincClientPtr cl, int type, unsigned int msg_len, 
		     int flags);
int thincWriteData(thincClientPtr cl, const void *buf, unsigned int count);
int thincBadWrite(thincClientPtr cl, const void *buf, unsigned int count);
int thincReadDataNoBuffer(thincClientPtr cl, void *buf, unsigned int count);
void* thincBufferMsg(thincClientPtr cl, int type, int flags, 
		     unsigned int msg_len, int data_len);
int thincClientRBufferEmpty(thincClientPtr cl);
int thincFlushBuffer(thincClientPtr cl);
int thincFlushBufferAll(thincScreenPtr pThinc);
int thincReadDataSocket(thincClientPtr cl, void *buf, unsigned int count);
int thincReadDataSecure(thincClientPtr cl, void *buf, unsigned int count);
int thincWriteDataSocket(thincClientPtr cl, const void *buf, 
			 unsigned int count);
int thincWriteDataSecure(thincClientPtr cl, const void *buf, 
			 unsigned int count);

#define T_TIMER	1000  /* ms, 1s */

OsTimerPtr thincInitATimer(thincScreenPtr pThinc);
int thincInitTimer(thincScreenPtr pThinc);
OsTimerPtr thincSetATimer(thincScreenPtr pThinc, OsTimerPtr timer, int flags, 
	int millis, OsTimerCallback func, pointer arg);
int thincSetTimer(thincScreenPtr pThinc, int flags, int millis, 
	OsTimerCallback func, pointer arg);
CARD32 thincTimerCallback(OsTimerPtr timer, CARD32 time, pointer arg);

void thincLoadExtension(thincScreenPtr pThinc);

int thincSStatsNew(thincScreenPtr pThinc);
inline struct thincScreenStats* thincSStatsFind(thincScreenPtr pThinc, 
	int stats_id);
inline void thincSStatsFree(thincScreenPtr pThinc, int stats_id);
inline void thincSStatsFreeIt(thincScreenPtr pThinc, struct thincScreenStats *);

int thincSetDebugLevel(thincScreenPtr pThinc, int debugLevel);
int thincGetDebugLevel(thincScreenPtr pThinc);

#if 0
int thincGetMouseDown(thincScreenPtr pThinc);
int thincGetMouseUp(thincScreenPtr pThinc);
int thincGetKeybDown(thincScreenPtr pThinc);
int thincGetKeybUp(thincScreenPtr pThinc);
int thincGetDisplayCount(thincScreenPtr pThinc);
int thincResetMouseDown(thincScreenPtr pThinc);
int thincResetMouseUp(thincScreenPtr pThinc);
int thincResetKeybDown(thincScreenPtr pThinc);
int thincResetKeybUp(thincScreenPtr pThinc);
int thincResetDisplayCount(thincScreenPtr pThinc);
#else
#define t_dec_inc_get_and_reset(name, value) \
inline void thincInc##name(thincScreenPtr pThinc); 		\
inline int thincGet##name(thincScreenPtr pThinc, int id); 	\
inline int thincReset##name(thincScreenPtr pThinc, int id);

t_dec_inc_get_and_reset(MouseDown, mouse_down)
t_dec_inc_get_and_reset(MouseUp, mouse_up)
t_dec_inc_get_and_reset(KeybDown, keyb_down)
t_dec_inc_get_and_reset(KeybUp, keyb_up)
t_dec_inc_get_and_reset(DisplayCount, display_count)
#endif

int thincGetScreensaver(thincScreenPtr pThinc);
void thincAddDisplayChange(thincScreenPtr pThinc, RegionPtr reg);
int thincGetDisplayPercent(thincScreenPtr pThinc, int id);
int thincResetDisplayPercent(thincScreenPtr pThinc, int id);

/**************************************************************************
 * 				Wrapper Functions 			  *
 **************************************************************************/

/* thincWindow.c */
#ifdef T_APPSHARE
Bool thincCreateWindow	(WindowPtr pWindow);
Bool thincDestroyWindow	(WindowPtr pWindow);
Bool thincPositionWindow(WindowPtr pWindow, int x, int y);
Bool thincChangeWindowAttributes(WindowPtr pWindow, unsigned long mask);
/* (Un)Realize Window */
Bool thincMapWindow	(WindowPtr pWindow);
Bool thincUnmapWindow	(WindowPtr pWindow);
void thincWindowExposures(WindowPtr pWindow, RegionPtr pRegion,
			 RegionPtr pBSRegion);
#endif	/* T_APPSHARE */
/* PaintWindowBackground/Border */
void thincPaintWindowBorder(WindowPtr pWindow, RegionPtr pRegion, int what);
void thincPaintWindowBackground(WindowPtr pWindow, RegionPtr pRegion, int what);
void thincCopyWindow	(WindowPtr pWindow, DDXPointRec ptOrigSrc, 
			 RegionPtr pRegionSrc);

/* thincPixmap.c */
PixmapPtr thincCreatePixmap(ScreenPtr pScreen,int width,int height, int depth);
Bool thincDestroyPixmap(PixmapPtr pPixmap);
Bool thincModifyPixmapHeader(PixmapPtr pPixmap, int width, int height,
	int depth, int bitsPerPixel, int devKind, pointer pPixData);

/* thincCursor.c */
int THINCHWCursorInit(ScreenPtr pScreen);
void THINCHWCursorUninit(thincScreenPtr pThinc);

/* thincGC.c */
Bool thincCreateGC	(GCPtr pGC);
void thincValidateGC	(GCPtr pGC, unsigned long changes, DrawablePtr pDraw);
RegionPtr thincCopyArea	(DrawablePtr dst, DrawablePtr src, GCPtr pGC,
			 int srcx, int srcy, int w, int h, int dstx, int dsty);
RegionPtr thincCopyPlane(DrawablePtr dst, DrawablePtr src, GCPtr pGC, 
			 int srcx, int srcy, int w, int h, int dstx, int dsty, 
			 unsigned long plane);
void thincPolyFillRect	(DrawablePtr dst, GCPtr pGC, int n, xRectangle *pRect);
void thincPutImage	(DrawablePtr dst, GCPtr pGC, int depth, int x, int y, 
			 int w, int h, int leftpad, int format, char *pBits);
void thincPolyPoint	(DrawablePtr dst, GCPtr pGC, int mode, int n, 
			 DDXPointPtr pPoint);
void thincPolylines	(DrawablePtr dst, GCPtr pGC, int mode, int n, 
			 DDXPointPtr pPoint);
void thincPolySegment	(DrawablePtr dst, GCPtr pGC, int n, 
			 xSegment *pSegments);
void thincPolyRectangle	(DrawablePtr dst, GCPtr pGC, int n, xRectangle *pRect);
void thincPolyArc	(DrawablePtr dst, GCPtr pGC, int n, xArc*pArc);
void thincFillPolygon	(DrawablePtr dst, GCPtr pGC,int shape, int mode, 
			 int count, DDXPointPtr pPoint);
void thincPolyFillArc	(DrawablePtr dst, GCPtr pGC, int n, xArc *pArc);
void thincImageText8	(DrawablePtr pDraw, GCPtr pGC, int x, int y, int count, 
			 char *chars);
int thincPolyText8	(DrawablePtr pDraw, GCPtr pGC, int x, int y, int count, 
			 char *chars);
void thincImageText16	(DrawablePtr pDraw, GCPtr pGC, int x, int y, int count,
			 unsigned short *chars);
int thincPolyText16	(DrawablePtr pDraw, GCPtr pGC, int x, int y, int count, 
			 unsigned short *chars);

/* The pixblit procedures (thincGC.c) */
void thincFillSpans	(DrawablePtr dst, GCPtr pGC, int nSpans, 
			 DDXPointPtr pPoints, int *pWidths, int sorted);
void thincSetSpans	(DrawablePtr pDrawable, GCPtr pGC, char *pSrc, 
			 DDXPointPtr pPoints, int *pWidths, int nSpans, 
			 int sorted);
void thincPolyGlyphBlt	(DrawablePtr pDrawable, GCPtr pGC, int x, int y, 
			 unsigned int nglyph, CharInfoPtr *ppci, 
			 pointer unused);
void thincImageGlyphBlt	(DrawablePtr pDrawable, GCPtr pGC, int x , int y, 
			 unsigned int nglyph, CharInfoPtr *ppci, 
			 pointer unused);
void thincPushPixels	(GCPtr pGC, PixmapPtr pBitMap, DrawablePtr pDrawable,
			 int dx, int dy, int xOrg, int yOrg);

#ifdef T_USEMITSHM
PixmapPtr thincShmCreatePixmap(ScreenPtr pScreen, int width,
	int height, int depth, char *addr);
void thincShmPutImage(DrawablePtr dst, GCPtr pGC, int depth,
	unsigned int format, int w, int h, int sx, int sy, int sw, int sh, 
	int dx, int dy, char *data);
#endif

/** Video **/
int thincInitVideo(thincScreenPtr pThinc);
int thincHandleVidStartOK(thincClientPtr cl);
#ifdef T_USEXVIDEO
int thincVInitNewClient(thincClientPtr cl);
int thincVideoDisable(thincClientPtr cl);
int thincVInitClientFmts(thincClientPtr cl, thinc_vidCInit *c);
void thincVidCloseClient(thincScreenPtr pThinc, thincClientPtr cl);
int thincVidClientResized(thincClientPtr cl, int prevstate);
#ifdef T_RECORD
int thincVInitRecordClient(thincClientPtr cl);
#endif /* T_RECORD */
int thincClientConnectedVideo(thincClientPtr cl);
#endif

/** Resize **/

/* depending on client's resize state (cl->resized), updates are handled 
 * different: */
#define 	T_CL_RESZ_NONE	0	/* not resized, do nothing */
#define		T_CL_RESZ_LESS	1	/* size less than fb size. resize
					   RAW updates */
#define 	T_CL_RESZ_MORE	2	/* size larger than fb size. do
					   nothing */
int thincHandleReszEvent(thincClientPtr cl);
int thincHandleReszUpdate(thincClientPtr cl);
void thincReszBox(thincClientPtr cl, BoxPtr rect);
int thinc_rawStackClResized(thincClientPtr cl, int prevstate);
int thinc_pixStackClResized(thincClientPtr cl, int prevstate);

/* stats and instrumentation.
 * thincInstrument.h doesn't know about pThinc so we leave these here */
void thincStatsInit(thincScreenPtr pThinc);
void thincStatsReset(thincScreenPtr pThinc);
void thincStatsPrint(thincScreenPtr pThinc);

int thincInstrInit(thincScreenPtr pThinc);
void thincInstrEnd(thincScreenPtr pThinc);
void thincInstrPrint(thincScreenPtr pThinc);
void thincInstrReset(thincScreenPtr pThinc);

/*** Maximum sizes used for some of the messages ***/
#define		T_MAXW	48
#define		T_MAXH 	48

/* Pixmap Fills have a limited size */
#define T_PIXMAP_MAXS	(T_MAXW*T_MAXH)

/* Max size of the bitmap sent in a Glyph or Bilevel command */
#define T_PIXELS_PER_BYTE	8
#define T_PIXELS_PER_LONG	32
#define T_LOG2_PIXELS_PER_BYTE	3
#define T_LOG2_PIXELS_PER_LONG	5
#if 0
#define T_MAX_BITMAP_SIZE	((T_MAXW/T_PIXELS_PER_BYTE)*T_MAXH)
#else
#define T_MAX_BITMAP_SIZE \
	(( (T_MAXW+T_PIXELS_PER_BYTE-1) >> T_LOG2_PIXELS_PER_BYTE)*T_MAXH)
#endif

/*** Additional #includes ***/

/* message stuff */
#include "thincMsg.h"

/* grayscale palette (needed by thincCommand.h) */
#include "thincPalette.h"

/* commands */
#include "thincCommand.h"

/* Offscreen stuff */
#include "thincPixmap.h"

/* Buffering stuff */
#include "thincBuffer.h"

/* window stuff */
#include "thincWindow.h"

#include "thincCompress.h"

/* message stack stuff */
#include "thincMsgStack.h"

/* authentication stuff */
#include "thinc_auth.h"

#ifdef T_USERENDER
#include "thincRender.h"
#endif

#include "thincBitFun.h"

/* application sharing */
#include "thincAppshare.h"

/* command recording */
#ifdef T_RECORD
#include "thincRecord.h"
#endif

#endif /* _THINC_H_ */
