/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * THINC Extension
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"
#include "dixstruct.h" /* for ClientPtr */
#include "extnsionst.h"
#include <X11/extensions/xthincstr.h>

static void THINCExtensionInit(void);
static int ProcTHINCDispatch(ClientPtr client);
static int SProcTHINCDispatch(ClientPtr client);
static void THINCResetProc(ExtensionEntry *extEntry);

static int thincErrorBase;
static int thincEventBase;

ExtensionModule thincExtensionModule = {
	THINCExtensionInit,	/* initFunc */
	"THINC",		/* name */
	NULL,			/* disablePtr */
	NULL,			/* setupFunc */
	NULL			/* initDependencies */
};

/* functions to handle extension requests */
#define t_procProto(name) 			\
static int ProcTHINC##name(ClientPtr client);	\
static int SProcTHINC##name(ClientPtr client);

t_procProto(QueryVersion)
t_procProto(GetID)
t_procProto(SetValue)
t_procProto(GetValue)
t_procProto(ResetValue)
t_procProto(StatsFlush)
t_procProto(StatsReset)
t_procProto(InstrFlush)
t_procProto(InstrReset)
t_procProto(WindowShare)

int (*ProcTHINCVector[XTHINCNumberRequests])(ClientPtr) = {
	ProcTHINCQueryVersion,
	ProcTHINCGetID,
	ProcTHINCSetValue,
	ProcTHINCGetValue,
	ProcTHINCResetValue,
	ProcTHINCStatsFlush,
	ProcTHINCStatsReset,
	ProcTHINCInstrFlush,
	ProcTHINCInstrReset,
	ProcTHINCWindowShare
};

int (*SProcTHINCVector[XTHINCNumberRequests])(ClientPtr) = {
	SProcTHINCQueryVersion,
	SProcTHINCGetID,
	SProcTHINCSetValue,
	SProcTHINCGetValue,
	SProcTHINCResetValue,
	SProcTHINCStatsFlush,
	SProcTHINCStatsReset,
	SProcTHINCInstrFlush,
	SProcTHINCInstrReset,
	SProcTHINCWindowShare
};

void thincLoadExtension(thincScreenPtr pThinc)
{
	static int first = T_TRUE;

	if(first) {
		/* FALSE -> builtin */
		LoadExtension(&thincExtensionModule, FALSE);
		first = T_FALSE;
	}
}

static void THINCExtensionInit(void)
{
	t_dbg_name(func)
	ExtensionEntry      *extEntry;

	thincDbg(T_DBG_OTHER, ("%s: initialiazing THINC extension\n", func));

	/*
	 * AddExtension(char *name, int NumEvents, int NumErrors, 
	 *              int (*MainProc)(ClientPtr c1), 
	 *              int (*SwappedMainProc)(ClientPtr c2), 
	 *              void (*CloseDownProc)(ExtensionEntry *e), 
	 *              unsigned short (*MinorOpcodeProc)(ClientPtr c3))
	 */

	extEntry = AddExtension(XTHINC_EXTENSION_NAME,
			xthincNumEvents, xthincNumErrors,
			ProcTHINCDispatch, SProcTHINCDispatch,
			THINCResetProc, StandardMinorOpcode);

	thincErrorBase = extEntry->errorBase;
	thincEventBase = extEntry->eventBase;

	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
}

/* Called whenever a client accesses the major opcode assigned to the extension. */
static int ProcTHINCDispatch(ClientPtr client)
{
	REQUEST(xReq);

	if(stuff->data >= XTHINCNumberRequests)
		return BadRequest;

	return (*ProcTHINCVector[stuff->data])(client);
}

/* Called whenever a client accesses the major opcode except the client using 
 * the extension has reversed byte-sex. */
static int SProcTHINCDispatch(ClientPtr client)
{
	REQUEST(xReq);

	if(stuff->data >= XTHINCNumberRequests)
		return BadRequest;

	return (*SProcTHINCVector[stuff->data])(client);
}

/* called at server reset time to deallocate any private storage used by the 
 * extension. */
static void THINCResetProc(ExtensionEntry *extEntry)
{
	/* noop */
}

/**************************************************************************
 *                    Extension Handling Functions                        *
 **************************************************************************/

#define t_reply_return(client, type, rep) 				\
	(void)WriteToClient(client, SIZEOF(type), (char *)&(rep));	\
	return ((client)->noClientException);

/* FIXME: If have more than one screen, would need to loop over
 * them, or make the client pass a screen identifier */
#define t_find_pThinc(client) 				\
	ScreenPtr pScreen = screenInfo.screens[0];	\
	thincScreenPtr pThinc = THINCPtr(pScreen)

/***** Query Version *****/
static int ProcTHINCQueryVersion(ClientPtr client)
{
	xTHINCQueryVersionReply rep;

	REQUEST_SIZE_MATCH(xTHINCQueryVersionReq);
	rep.type 		= X_Reply;
	rep.sequenceNumber  	= client->sequence;
	rep.length		= 0;
	/* TODO: Returning extension version number for now. should we add
	 * a new call to query the protocol version? that one needs 3 
	 * numbers...
	 */
	rep.majorVersion	= XTHINC_VERSION_MAJOR;
	rep.minorVersion	= XTHINC_VERSION_MINOR;
	if(client->swapped) {
		register char n;
		swaps(&rep.sequenceNumber, n);
		swaps(&rep.length, n);
		swaps(&rep.majorVersion, n);
		swaps(&rep.minorVersion, n);
	}
	t_reply_return(client, xTHINCQueryVersionReply, rep);
}

static int ProcTHINCGetID(ClientPtr client)
{
	xTHINCGetIDReply rep;
	t_find_pThinc(client);

	REQUEST_SIZE_MATCH(xTHINCGetIDReq);
	rep.type 		= X_Reply;
	rep.sequenceNumber  	= client->sequence;
	rep.length		= 0;
	rep.id			= thincSStatsNew(pThinc);
	if(rep.id == T_ERROR) 
		return BadAlloc;

	if(client->swapped) {
		register char n;
		swaps(&rep.sequenceNumber, n);
		swaps(&rep.length, n);
		swaps(&rep.id, n);
	}
	t_reply_return(client, xTHINCGetIDReply, rep);
}

static int ProcTHINCSetValue(ClientPtr client)
{
	t_dbg_name(func)
	int err = T_ERROR;	/* to avoid compiler bitching */
	xTHINCSetValueReply rep;
	t_find_pThinc(client);
	REQUEST(xTHINCSetValueReq);
	REQUEST_SIZE_MATCH(xTHINCSetValueReq);

	thincDbg(T_DBG_OTHER, ("%s: setting value %ld to %ld\n", func,
		stuff->which, stuff->value));

	switch(stuff->which) {
		case xTHINC_val_debugLevel:
			err = thincSetDebugLevel(pThinc, stuff->value);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unknown "
				"value %ld to set\n", func, stuff->which));
			return BadRequest;
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't set value "
			"%ld to %ld\n", func, stuff->which, stuff->value));
		return BadValue;
	}

	/* set up the reply and off we go */
	rep.type 		= X_Reply;
	rep.sequenceNumber  	= client->sequence;
	rep.length		= 0;
	rep.value		= stuff->value; /* kind of pointless :) */
	if(client->swapped) {
		register char n;
		swaps(&rep.sequenceNumber, n);
		swaps(&rep.length, n);
		swaps(&rep.value, n);
	}
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	t_reply_return(client, xTHINCSetValueReply, rep);
}

static int ProcTHINCGetValue(ClientPtr client)
{
	t_dbg_name(func)
	int value = T_ERROR;	/* to avoid compiler bitching */
	xTHINCGetValueReply rep;
	t_find_pThinc(client);

	REQUEST(xTHINCGetValueReq);
	REQUEST_SIZE_MATCH(xTHINCGetValueReq);

	thincDbg(T_DBG_OTHER, ("%s: getting value %ld\n", func,
		stuff->which));

	switch(stuff->which) {
		case xTHINC_val_debugLevel:
			value = thincGetDebugLevel(pThinc);
			break;
		case xTHINC_val_mouseDown:
			value = thincGetMouseDown(pThinc, stuff->id);
			break;
		case xTHINC_val_mouseUp:
			value = thincGetMouseUp(pThinc, stuff->id);
			break;
		case xTHINC_val_keybDown:
			value = thincGetKeybDown(pThinc, stuff->id);
			break;
		case xTHINC_val_keybUp:
			value = thincGetKeybUp(pThinc, stuff->id);
			break;
		case xTHINC_val_screensaver:
			value = thincGetScreensaver(pThinc);
			break;
		case xTHINC_val_displayCount:
			value = thincGetDisplayCount(pThinc, stuff->id);
			break;
		case xTHINC_val_displayPercent:
			value = thincGetDisplayPercent(pThinc, stuff->id);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unknown "
				"value %ld to get\n", func, stuff->which));
			return BadRequest;
	}
	if(value == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't get value "
			"%ld for id %ld\n", func, stuff->which, stuff->id));
		return BadValue;
	}

	/* set up the reply and off we go */
	rep.type 		= X_Reply;
	rep.sequenceNumber  	= client->sequence;
	rep.length		= 0;
	rep.which		= stuff->which;
	rep.value		= value;
	if(client->swapped) {
		register char n;
		swaps(&rep.sequenceNumber, n);
		swaps(&rep.length, n);
		swaps(&rep.which, n);
		swaps(&rep.value, n);
	}
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	t_reply_return(client, xTHINCGetValueReply, rep);
}

static int ProcTHINCResetValue(ClientPtr client)
{
	t_dbg_name(func)
	int err = T_ERROR;	/* to avoid compiler bitching */
	xTHINCResetValueReply rep;
	t_find_pThinc(client);

	REQUEST(xTHINCResetValueReq);
	REQUEST_SIZE_MATCH(xTHINCResetValueReq);

	thincDbg(T_DBG_OTHER, ("%s: resetting value %ld\n", func,
		stuff->which));

	switch(stuff->which) {
		case xTHINC_val_mouseDown:
			err = thincResetMouseDown(pThinc, stuff->id);
			break;
		case xTHINC_val_mouseUp:
			err = thincResetMouseUp(pThinc, stuff->id);
			break;
		case xTHINC_val_keybDown:
			err = thincResetKeybDown(pThinc, stuff->id);
			break;
		case xTHINC_val_keybUp:
			err = thincResetKeybUp(pThinc, stuff->id);
			break;
		case xTHINC_val_displayCount:
			err = thincResetDisplayCount(pThinc, stuff->id);
			break;
		case xTHINC_val_displayPercent:
			err = thincResetDisplayPercent(pThinc, stuff->id);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unknown "
				"value %ld to reset\n", func, stuff->which));
			return BadRequest;
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't reset value "
			"%ld for id %ld\n", func, stuff->which, stuff->id));
		return BadValue;
	}

	/* set up the reply and off we go */
	rep.type 		= X_Reply;
	rep.sequenceNumber  	= client->sequence;
	rep.length		= 0;
	if(client->swapped) {
		register char n;
		swaps(&rep.sequenceNumber, n);
		swaps(&rep.length, n);
	}
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	t_reply_return(client, xTHINCResetValueReply, rep);
}

/* write current stats to log file */
static int ProcTHINCStatsFlush(ClientPtr client)
{
	t_dbg_name(func)
	t_find_pThinc(client);

	REQUEST_SIZE_MATCH(xTHINCStatsFlushReq);

	thincDbg(T_DBG_OTHER, ("%s: flushing stats\n", func));
	thincStatsPrint(pThinc);
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	return client->noClientException;
}

static int ProcTHINCStatsReset(ClientPtr client)
{
	t_dbg_name(func)
	t_find_pThinc(client);

	REQUEST_SIZE_MATCH(xTHINCStatsFlushReq);

	thincDbg(T_DBG_OTHER, ("%s: resetting stats\n", func));
	thincStatsReset(pThinc);
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	return client->noClientException;
}

static int ProcTHINCInstrFlush(ClientPtr client)
{
	t_dbg_name(func)
	t_find_pThinc(client);

	REQUEST_SIZE_MATCH(xTHINCInstrFlushReq);

	thincDbg(T_DBG_OTHER, ("%s: flushing instrumentation\n", func));
	thincInstrPrint(pThinc);
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	return client->noClientException;
}

static int ProcTHINCInstrReset(ClientPtr client)
{
	t_dbg_name(func)
	t_find_pThinc(client);

	REQUEST_SIZE_MATCH(xTHINCInstrFlushReq);

	thincDbg(T_DBG_OTHER, ("%s: resetting instrumentation\n", func));
	thincInstrReset(pThinc);
	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	return client->noClientException;
}

static int ProcTHINCWindowShare(ClientPtr client)
{
#ifdef T_APPSHARE
	t_dbg_name(func)
	WindowPtr pWindow;
	int status;
	t_find_pThinc(client);

	REQUEST(xTHINCWindowShareReq);
	REQUEST_SIZE_MATCH(xTHINCWindowShareReq);

#if 0	/* function has been deprecated */
	pWindow = SecurityLookupWindow(stuff->window, client,
		SecurityReadAccess);
#else
	status = dixLookupWindow(&pWindow, stuff->window, client, 
		DixReadAccess);
#endif
	if(status != Success) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: invalid window %#lx to "
			"share\n", func, stuff->window));
		return status;
	}

	thincDbg(T_DBG_OTHER, ("%s: sharing window (%#lx) %#lx\n", func, 
		stuff->window, t_winID(pWindow)));

	if(thincWindowShare(pThinc, pWindow) == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't share window "
			"%#lx\n", func, t_winID(pWindow)));
		/* sketchy return value as probably the only reason we 
		 * can fail here is a network error */
		return BadWindow;
	}

	thincDbg(T_DBG_OTHER, ("%s: finished\n", func));
	return client->noClientException;
#else
	return BadRequest;
#endif
}

/**** SProc functions. kind of the same as Proc funcs ****/
static int SProcTHINCQueryVersion(ClientPtr client)
{
	REQUEST(xTHINCQueryVersionReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCQueryVersionReq);
	swaps(&stuff->majorVersion, n);
	swaps(&stuff->minorVersion,n);
	return ProcTHINCQueryVersion(client);
}

static int SProcTHINCGetID(ClientPtr client)
{
	REQUEST(xTHINCGetIDReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCGetIDReq);
	return ProcTHINCGetID(client);
}

static int SProcTHINCSetValue(ClientPtr client)
{
	REQUEST(xTHINCSetValueReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCSetValueReq);
	swaps(&stuff->which,n);
	swaps(&stuff->value,n);
	return ProcTHINCSetValue(client);
}

static int SProcTHINCGetValue(ClientPtr client)
{
	REQUEST(xTHINCGetValueReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCGetValueReq);
	swaps(&stuff->which, n);
	return ProcTHINCGetValue(client);
}

static int SProcTHINCResetValue(ClientPtr client)
{
	REQUEST(xTHINCGetValueReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCResetValueReq);
	swaps(&stuff->which, n);
	return ProcTHINCResetValue(client);
}

static int SProcTHINCStatsFlush(ClientPtr client)
{
	REQUEST(xTHINCStatsFlushReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCStatsFlushReq);
	return ProcTHINCStatsFlush(client);
}

static int SProcTHINCStatsReset(ClientPtr client)
{
	REQUEST(xTHINCStatsResetReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCStatsResetReq);
	return ProcTHINCStatsReset(client);
}

static int SProcTHINCInstrFlush(ClientPtr client)
{
	REQUEST(xTHINCInstrFlushReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCInstrFlushReq);
	return ProcTHINCInstrFlush(client);
}

static int SProcTHINCInstrReset(ClientPtr client)
{
	REQUEST(xTHINCInstrResetReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCInstrResetReq);
	return ProcTHINCInstrReset(client);
}

static int SProcTHINCWindowShare(ClientPtr client)
{
	REQUEST(xTHINCWindowShareReq);
	register char n;

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xTHINCWindowShareReq);
	return ProcTHINCWindowShare(client);
}
