/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/*
 * Helper macros for GC functions
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

extern GCOps thincGCOps;

/* processing common to all GC funcs */
#define thincGC_pre(pGC) \
	ScreenPtr pScreen = pGC->pScreen; \
	thincScreenPtr pThinc = THINCPtr(pScreen); \
	thincGCPtr pThincGC = THINCGCPtr(pGC); \
	/* unwrap ops vector so we don't send update twice */ \
	pGC->ops = pThincGC->driver_ops;

#define thincGC_post(pGC) \
	/* wrap ops vector again */ \
	pGC->ops = &thincGCOps;

/* helper macro to decide if target drawable is visible and we should
 * send an update to the client */
#define should_update(dst) do { \
	if ((dst)->type == DRAWABLE_WINDOW &&	\
		((WindowPtr) (dst))->realized) 	\
		send_update = 1; \
	else {	\
		thincDbg(T_DBG_DRAW, \
			("%s: Not updating pixmap with SN %lx\n", func, \
			(dst)->serialNumber)); \
		send_update = 0; \
	} \
} while(0)

/* TODO: We could also put unrealized windows (unviewable may imply that the
 * window is InputOnly) */
#define is_offscreen(pDraw) \
	((pDraw)->type == DRAWABLE_PIXMAP)

#define is_onscreen(pDraw) \
	(((pDraw)->type == DRAWABLE_WINDOW) && ((WindowPtr) (pDraw))->viewable)

/* prints useful information about a drawable. 
 * Be careful not to print more than 255 chars!!! */
#define thincDrawable_print(pDraw, str) do { \
	char prtstring[256];	\
	char *curr = prtstring; \
	int count; \
	if((pDraw)->type == DRAWABLE_WINDOW) { \
		count = sprintf(curr, "id: %#lx. Drawable Window. ", \
			(pDraw)->id); \
		curr += count; \
		if(((WindowPtr) (pDraw))->realized) \
			count = sprintf(curr, "Realized."); \
		else \
			count = sprintf(curr, "Unrealized."); \
		curr += count; \
	} else if((pDraw)->type == DRAWABLE_PIXMAP) { \
		count = sprintf(curr, " id: %#lx. Drawable Pixmap. SN %#lx", \
			t_pixID((PixmapPtr)pDraw), (pDraw)->serialNumber); \
		curr += count;	\
	} else if((pDraw)->type == UNDRAWABLE_WINDOW) { \
		count = sprintf(curr, "id: %#lx. Undrawable Window.", \
			(pDraw)->id); \
		curr += count;	\
	} else {	\
		count = sprintf(curr, "id: %#lx. Unknown type %d", \
			(pDraw)->id, (pDraw)->type); \
		curr += count; \
	} \
	thincDbg(T_DBG_DRAW, ("%s: %s%s\n", func, str, prtstring)); \
} while(0)

int thincGC_clipUpdate(GCPtr gc, RegionPtr reg, int x, int y, int w, int h);
int thincGC_clipUpdateRegion(GCPtr gc, RegionPtr update);

/* macros to call clipUpdate and verify the return value */
#define call_clipUpdate(pGC, reg, x, y, w, h, ret) \
	if(thincGC_clipUpdate(pGC, reg, x, y, w, h) == T_ERROR) { \
		thincDbg(T_DBG_ERROR, \
			("ERROR: %s: Couldn't clip update region\n", func)); \
		thincGC_post(pGC); \
		return ret; \
	}

#define call_clipUpdateRegion(pGC, reg, ret) \
	if(thincGC_clipUpdateRegion(pGC, reg) == T_ERROR) { \
		thincDbg(T_DBG_ERROR, \
			("ERROR: %s: Couldn't clip update region\n", func)); \
		REGION_DESTROY(pScreen, reg); \
		thincGC_post(pGC); \
		return ret; \
	}

#define punt_to_raw(pThinc, isOff, res, dst, reg) do { \
	if(isOff) \
		res = thincQueueCommand(pThinc, (PixmapPtr)dst, T_FB_RUPDATE, \
			reg, T_RAW_VA_NONE); \
	else \
		res = thincBufferNewCmdReg(pThinc, (WindowPtr)dst, \
			T_FB_RUPDATE, reg, T_RAW_VA_NONE); \
} while(0)

/* use when you don't want to return anything */
#define noret 
