/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Functions for statistics
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

/*************** stats ******************/

void thincStatsInit(thincScreenPtr pThinc)
{
	/* per command */
	stats_str_init(&pThinc->rawStats);
	stats_str_init(&pThinc->copyStats);
	stats_str_init(&pThinc->sfillStats);
	stats_str_init(&pThinc->pfillStats);
	stats_str_init(&pThinc->glyphStats);
	stats_str_init(&pThinc->bilevelStats);

	/* draw stats */
	stats_init(pThinc, countFillSpans);
	stats_init(pThinc, countSetSpans);
	stats_init(pThinc, countPutImage);
	stats_init(pThinc, countCopyArea);
	stats_init(pThinc, countCopyPlane);
	stats_init(pThinc, countPolyPoint);
	stats_init(pThinc, countPolylines);
	stats_init(pThinc, countPolySegment);
	stats_init(pThinc, countPolyRectangle);
	stats_init(pThinc, countPolyArc);
	stats_init(pThinc, countFillPolygon);
	stats_init(pThinc, countPolyFillRect);
	stats_init(pThinc, countPolyFillArc);
	stats_init(pThinc, countPolyText8);
	stats_init(pThinc, countPolyText16);
	stats_init(pThinc, countImageText8);
	stats_init(pThinc, countImageText16);
	stats_init(pThinc, countImageGlyphBlt);
	stats_init(pThinc, countPolyGlyphBlt);
	stats_init(pThinc, countPushPixels);
	stats_init(pThinc, countPaintWindowBorder);
	stats_init(pThinc, countPaintWindowBackground);
	stats_init(pThinc, countCopyWindow);

	stats_init(pThinc, merge_try);
	stats_init(pThinc, merges);
	stats_init(pThinc, merge_raw);
	stats_init(pThinc, merge_sfill);
	stats_init(pThinc, merge_time);
	stats_init(pThinc, merge_time_alloc);
	stats_init(pThinc, merge_time_free);

	stats_init(pThinc, overwrite_try);
	stats_init(pThinc, overwrite_try_cmds);
	stats_init(pThinc, overwrite_any);
	stats_init(pThinc, overwrite_any_cmds);
	stats_init(pThinc, overwrites);
	stats_init(pThinc, overwrite_complete);
	stats_init(pThinc, overwrite_noover);
	stats_init(pThinc, overwrite_trans);
	stats_init(pThinc, overwrite_time);

	stats_init(pThinc, find_deps);

	stats_init(pThinc, overwrite_type[T_FB_RUPDATE]);
	stats_init(pThinc, overwrite_type[T_FB_COPY]);
	stats_init(pThinc, overwrite_type[T_FB_SFILL]);
	stats_init(pThinc, overwrite_type[T_FB_PFILL]);
	stats_init(pThinc, overwrite_type[T_FB_GLYPH]);
	stats_init(pThinc, overwrite_type[T_FB_BILEVEL]);
	stats_init(pThinc, overwrite_type[T_FB_CURSOR]);
}

void thincStatsReset(thincScreenPtr pThinc)
{
	thincStatsInit(pThinc);
}

void thincStatsPrint(thincScreenPtr pThinc)
{
	stats_log(("::: Statistics :::\n"));
	stats_log(("[NOTE: sizes are computed as header + message + data]\n"));

	/* per command */
	stats_log(("RAW:\n"));
	stats_log(("[Size of message: %d bytes]\n", 
		THINC_HEADERSIZE + T_FB_RUPDATE_SIZE));
	stats_log(("[Size of compressed message: %d bytes]\n",
		T_FB_ZDATA_SIZE));
	stats_str_print(&pThinc->rawStats);
	stats_log(("COPY:\n"));
	stats_log(("[Size of message: %d bytes]\n", 
		THINC_HEADERSIZE + T_FB_COPY_SIZE));
	stats_str_print(&pThinc->copyStats);
	stats_log(("SFILL:\n"));
	stats_log(("[Size of message: %d bytes]\n", 
		THINC_HEADERSIZE + T_FB_SFILL_SIZE));
	stats_str_print(&pThinc->sfillStats);
	stats_log(("PFILL:\n"));
	stats_log(("[Size of message: %d bytes]\n", 
		THINC_HEADERSIZE + T_FB_PFILL_SIZE));
	stats_str_print(&pThinc->pfillStats);
	stats_log(("GLYPH:\n"));
	stats_log(("[Size of message: %d bytes]\n", 
		THINC_HEADERSIZE + T_FB_GLYPH_SIZE));
	stats_str_print(&pThinc->glyphStats);
	stats_log(("BILEVEL:\n"));
	stats_log(("[Size of message: %d bytes]\n", 
		THINC_HEADERSIZE + T_FB_BILEVEL_SIZE));
	stats_str_print(&pThinc->bilevelStats);

	/* draw stats */
	stats_log(("DRAW:\n"));
	stats_print(pThinc, countFillSpans);
	stats_print(pThinc, countSetSpans);
	stats_print(pThinc, countPutImage);
	stats_print(pThinc, countCopyArea);
	stats_print(pThinc, countCopyPlane);
	stats_print(pThinc, countPolyPoint);
	stats_print(pThinc, countPolylines);
	stats_print(pThinc, countPolySegment);
	stats_print(pThinc, countPolyRectangle);
	stats_print(pThinc, countPolyArc);
	stats_print(pThinc, countFillPolygon);
	stats_print(pThinc, countPolyFillRect);
	stats_print(pThinc, countPolyFillArc);
	stats_print(pThinc, countPolyText8);
	stats_print(pThinc, countPolyText16);
	stats_print(pThinc, countImageText8);
	stats_print(pThinc, countImageText16);
	stats_print(pThinc, countImageGlyphBlt);
	stats_print(pThinc, countPolyGlyphBlt);
	stats_print(pThinc, countPushPixels);
	stats_print(pThinc, countPaintWindowBorder);
	stats_print(pThinc, countPaintWindowBackground);
	stats_print(pThinc, countCopyWindow);
	/* TODO: Put this in STATS */
	stats_print(pThinc, droppedFrames);

	stats_print(pThinc, merge_try);
	stats_print(pThinc, merges);
	stats_print(pThinc, merge_raw);
	stats_print(pThinc, merge_sfill);
	stats_print(pThinc, merge_time);
	stats_print(pThinc, merge_time_alloc);
	stats_print(pThinc, merge_time_free);

	stats_print(pThinc, overwrite_try);
	stats_print(pThinc, overwrite_try_cmds);
	stats_print(pThinc, overwrite_any);
	stats_print(pThinc, overwrite_any_cmds);
	stats_print(pThinc, overwrites);
	stats_print(pThinc, overwrite_complete);
	stats_print(pThinc, overwrite_noover);
	stats_print(pThinc, overwrite_trans);
	stats_print(pThinc, overwrite_time);

	stats_print(pThinc, find_deps);

	stats_print(pThinc, overwrite_type[T_FB_RUPDATE]);
	stats_print(pThinc, overwrite_type[T_FB_COPY]);
	stats_print(pThinc, overwrite_type[T_FB_SFILL]);
	stats_print(pThinc, overwrite_type[T_FB_PFILL]);
	stats_print(pThinc, overwrite_type[T_FB_GLYPH]);
	stats_print(pThinc, overwrite_type[T_FB_BILEVEL]);
	stats_print(pThinc, overwrite_type[T_FB_CURSOR]);
}

#ifdef T_GETSTATS
void stats_str_init(struct thincStats *stats)
{
	stats_init(stats, count);
	stats_init(stats, zcount);
	stats_init(stats, size);
	stats_init(stats, zsize);
	stats_init(stats, create);
	stats_init(stats, destroy);
	stats_init(stats, copy);
	stats_init(stats, modify);
	stats_init(stats, info);
	stats_init(stats, exec);
	stats_init(stats, flush);
	stats_init(stats, write);
	stats_init(stats, block);
	stats_init(stats, total);
}

void stats_str_print(struct thincStats *stats)
{
	stats_log(("=> count: %lu\n", stats->count));
	stats_log(("=> size: %lu bytes\n", stats->size));
	if(stats->zsize) {
		stats_log(("=> compressed count: %lu\n", stats->zcount));
		stats_log(("=> compressed size: %lu bytes\n", stats->zsize));
	}
	stats->total = 0;
	stats_log(("=> create:  %6lums, destroy: %6lums, copy:  %6lums\n", 
		stats->create, stats->destroy, stats->copy));
	stats->total += stats->create;
	stats->total += stats->destroy;
	stats->total += stats->copy;
	stats_log(("=> modify:  %6lums, info:    %6lums, exec:  %6lums\n", 
		stats->modify, stats->info, stats->exec));
	stats->total += stats->modify;
	stats->total += stats->info;
	stats->total += stats->exec;
	stats_log(("=> flush:   %6lums, write:   %6lums, block: %6lums\n", 
		stats->flush, stats->write, stats->block));
	stats->total += stats->flush;
	stats->total += stats->write;
	stats->total += stats->block;
	stats_log(("=> total:   %6lums\n", stats->total));
}
#endif	/* T_GETSTATS */

/*************** instrumentation ******************/
#ifdef T_INSTRUMENT
struct t_irecord* t_ievents = NULL;
unsigned long t_icount = 0;
unsigned long t_iwrap = 0;

int thincInstrInit(thincScreenPtr pThinc)
{
	t_ievents = malloc((T_INSTR_HOWMANY)*sizeof(*t_ievents));
	if(!t_ievents)
		return T_ERROR;

	return T_OK;
}

void thincInstrEnd(thincScreenPtr pThinc)
{
	free(t_ievents);
	t_icount = t_iwrap = 0;
}

void thincInstrPrint(thincScreenPtr pThinc)
{
	register unsigned long i = 0;
	static int count = 0;

	instr_log(("EVENTS %d\n", count));
	instr_print_ev(TI_START);
	instr_print_ev(TI_END);
	instr_print_ev(TI_DRAW_START);
	instr_print_ev(TI_DRAW_END);
	instr_print_ev(TI_ALLOC_START);
	instr_print_ev(TI_ALLOC_END);
	instr_print_ev(TI_BUFFER_START);
	instr_print_ev(TI_BUFFER_END);
	instr_print_ev(TI_EXEC_QUEUE_START);
	instr_print_ev(TI_EXEC_QUEUE_END);
	instr_print_ev(TI_COPY_QUEUE_START);
	instr_print_ev(TI_COPY_QUEUE_END);
	instr_log(("EVENTS END\n"));

	instr_log(("TIMELINE %lu %lu\n", t_icount, t_iwrap));
	for( ; i<t_icount; ++i) {
		instr_print_rec(i);
	}
	instr_log(("TIMELINE END %d\n", count));
	++count;
}

void thincInstrReset(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	instr_dbg(("%s: resetting instrumentation. had %lu records\n", 
		func, t_icount));
	t_icount = t_iwrap = 0;
}
#else						/* T_INSTRUMENT */
int thincInstrInit(thincScreenPtr pThinc)
{
	return T_OK;
}

void thincInstrEnd(thincScreenPtr pThinc)
{
	/* noop */
}

void thincInstrPrint(thincScreenPtr pThinc)
{
	/* noop */
}

void thincInstrReset(thincScreenPtr pThinc)
{
	/* noop */
}

#endif 						/* T_INSTRUMENT */
