/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * an implementation of a simple grayscale color palette
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

static unsigned char thincPaletteReallyAdd(thincPalette *plte, 
	thincColor *color);

thincPalette *thincPaletteNew()
{
	t_dbg_name(func)
	thincPalette *plte;

	/* we alloc all the entries at once */
	plte = malloc(sizeof(*plte) + (sizeof(thincPlteEntry)*T_PLTE_MAX_SZ));
	if(!plte) {
		thincDbg(T_DBG_ERROR, ("ERRROR: %s: Out of memory?\n", func));
		return NULL;
	}

	plte->entries = 0;
	plte->plte = (thincPlteEntry*) (plte+1);

	memset(plte->plte, 0, (sizeof(thincPlteEntry)*T_PLTE_MAX_SZ));

	return plte;
}

void thincPaletteDestroy(thincPalette *plte)
{
	plte->plte = NULL;
	free(plte);
}

unsigned char thincPaletteAdd(thincPalette *plte, unsigned long color)
{
	thincPlteEntry *p = plte->plte;
	unsigned char idx = t_plte_idx(color);

	if(p[idx].active)
		return p[idx].idx;
	else {
		thincColor c;

		c.red = (color & 0xff0000) >> 16;
		c.green = (color & 0x00ff00) >> 8;
		c.blue = color & 0x0000ff;

		return thincPaletteReallyAdd(plte, &c);
	}
}

unsigned char thincPaletteAddColor(thincPalette *plte, thincColor *color)
{
	unsigned char idx = color->blue;	/* kinda cheating... */
	thincPlteEntry *p = plte->plte;

	if(p[idx].active)
		/* gots it, done */
		return p[idx].idx;
	
	return thincPaletteReallyAdd(plte, color);
}

/* assumes color is not in the palette. it will blindly overwrite
 * whatever is there */
static unsigned char thincPaletteReallyAdd(thincPalette *plte, thincColor *color)
{
	unsigned char idx = color->blue;	/* kinda cheating... */
	thincPlteEntry *p = plte->plte, *e;

	e = &p[idx];
	e->active = 1;
	e->idx = plte->entries;
	e->color.red = color->red;
	e->color.blue = color->blue;
	e->color.green = color->green;

	plte->entries++;

	return e->idx;
}

/* get all colors in the palette and add them to an array in the order
 * according to each color index. basically needed to converts to the format 
 * used by RAW (and PNG). returns the number of entries used */
unsigned int thincPaletteGetColors(thincPalette *plte, thincColor *colors)
{
	thincPlteEntry *e;
	thincColor *f, *t;
	unsigned int i;

	t_plte_for_each(plte, e, i) {
		if(!t_plte_active(e))
			continue;

		t = &colors[t_plte_index(e)];
		f = t_plte_color(e);
		t->red = f->red;
		t->green = f->green;
		t->blue = f->blue;
	}

	return plte->entries;
}
