/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/*
 * an implementation of a simple grayscale color palette
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#ifndef _THINC_PALETTE_H_
#define _THINC_PALETTE_H_

/* TODO: would it be ok to use png's definitions here instead?? */
typedef struct thincColor_S {
	unsigned char red;
	unsigned char green;
	unsigned char blue;
} thincColor;

typedef struct thincPlteEntry_S {
	int active:1;		/* do we have valid info? */
	int mbz:15;
	unsigned short idx;	/* the order at which this entry was added
				   to the palette */
	thincColor color;
} thincPlteEntry;

typedef struct thincPalette {
	thincPlteEntry *plte;
	unsigned int entries;
} thincPalette;

#define t_plte_entries(plte)	(plte)->entries

/* for now, all palettes have same maximum size (made for 8bpp color) */
#define T_PLTE_BIT_SZ	8U
#define T_PLTE_MAX_SZ	(1U<<T_PLTE_BIT_SZ)	/* 256 */
#define T_PLTE_MASK	(T_PLTE_MAX_SZ-1)	/* 0xff in 8bbp */

#define t_plte_idx(color)	(color & T_PLTE_MASK)

thincPalette *thincPaletteNew(void);
void thincPaletteDestroy(thincPalette *plte);

unsigned char thincPaletteAdd(thincPalette *plte, unsigned long color);
unsigned char thincPaletteAddColor(thincPalette *plte, thincColor *color);
/* get all colors in the palette and add them to an array in the order
 * according to each color index. basically needed to converts to the format 
 * used by RAW (and PNG). returns the number of entries used.
 * To convert from thincPalette to rawPalette call like this:
 * rplte->entries = thincPaletteGetColors(plte, rplte->palette);
 */
unsigned int thincPaletteGetColors(thincPalette *plte, thincColor *colors);

/* loop over the entries in the palette. i is int used as auxiliary counter.
 * you'll need to call t_plte_active(e) if you just want to loop over
 * the active entries */
#define t_plte_for_each(plte, e, i)	\
	for(i = 0, e = &plte->plte[0]; i<T_PLTE_MAX_SZ; ++i, ++e)

#define t_plte_active(e)	(e)->active
#define t_plte_index(e)		(e)->idx
#define t_plte_color(e)		&(e)->color

#endif /* _THINC_PALETTE_H_ */
