/* Copyright 2006, 2007 Ricardo Baratto */
/* Copyright 2006, 2007 Columbia University */
/*
 * $Id: $
 * $Source: $
 *
 * thincRecord.h: implements a command recorder
 *
 * Author:
 * Ricardo Baratto <ricardo@cs.columbia.edu>
 * Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

/*
 * TODO
 * > we should probably use off_t for all the file offsets, but htonl()'ing
 * them would be such a pain :(
 */

#ifndef _THINC_RECORD_H_
#define _THINC_RECORD_H_

#include "thincRecordCommon.h"

#define T_REC_DEFAULT_RATE	0	/* ms */

/* some flags */
#define T_REC_ALWAYS		0x0001

/* get record pointer from pThinc */
#define t_recordPtr(pThinc) (pThinc)->record

/* write 64K at a time */
#define T_REC_BUFSIZE		(1U<<16)

typedef struct thincRecord_S {
	thincScreenPtr 		pThinc;
	
	/* memory buffer */
	char 			*buf;
	char			*c_buf;
	unsigned int 		buf_size;

	/* len is without \0 */
	char*			dir;
	unsigned int		dir_len;

	/* file info */
	int			index;

	int			cmd;
	char			cmd_fn[T_REC_FILEID_LEN];
	int			cmd_fp;
	int			shot;
	char			shot_fn[T_REC_FILEID_LEN];
	int			shot_fp;

	/* current state */
	struct thincRecCmd	c_cmd;	/* current command */

	/* recorder client */
	thincClientPtr		cl;
	
	/* to take screenshots */
	thincMsgStackPtr	shot_stack;
	OsTimerPtr		shot_timer;

	/* time in ms (from gettime_ms()). when presenting to user have
	 * to convert to seconds and cast to time_t. then use localtime()
	 * or somethin... */
	unsigned long long		time;

	/* control behavior */
	unsigned int		always:1;
	unsigned int		cl_enabled:1;
	unsigned int		pad:31;

	thincListRec		state;

#if 0
	/* to optimize screenshot taking. can't reuse the one in pThinc
	 * cuz oreno's already using that one for his daemon */
	int			display_count;
	RegionRec		display_change;
#else
	int		stats_id;
#endif
} thincRecordRec, *thincRecordPtr;

typedef struct thincRecCmd_S {
	unsigned long		id;
	thincListRec		list;
	char			file[T_REC_FILEID_LEN];
	int			fp;
	RegionRec		reg;
} thincRecCmdRec, *thincRecCmdPtr;

int thincRecordInit(thincScreenPtr pThinc, const char *base_dir, int flags);
void thincRecordDestroy(thincScreenPtr pThinc);
int thincRecord(thincClientPtr cl, unsigned int len, int type, int flags,
	int (*rec_cb)(thincClientPtr, char *, unsigned int*, void *),
	void *handle);
int thincRecordInitState(thincScreenPtr pThinc);
void thincRecordDestroyState(thincScreenPtr pThinc);

#define T_REC_DONE	T_DONE
#define T_REC_MORE	T_PART

void thincRecordClientDisable(thincScreenPtr pThinc);
int thincRecordClientEnable(thincScreenPtr pThinc);

/* call to flush record buffer */
int thincRecordFlush(thincScreenPtr pThinc);

/* extra recording functions */
int thinc_shotInitStack(thincRecordPtr record);
void thinc_shotDestroyStack(thincRecordPtr record);
int thincRecordTakeShot(thincScreenPtr pThinc, int force);
int thincRecordGetNewShotFile(thincRecordPtr record);

#if 0
#define T_REC_SHOT_TIMER	600000 /* in ms, 10 minutes */
#else
#define T_REC_SHOT_TIMER	60000 /* in ms, 1 minute */
#endif

#if 0
/* max size of a file (leave it at 32bits tho since we're using off_t
 * should be able to handle more) */
#define T_REC_FPOS_SIZE	((sizeof(long)<<3) - 1)
#define T_REC_CMDFILE_MAXSZ	((1U << T_REC_FPOS_SIZE) - 1)
#else
/* max size of a file (one above may be too big?). note that this is by
 * no means a hard limit, just a hint that if we're past this size we
 * should probably create a new file */
#define T_REC_CMDFILE_MAXSZ	(1U << 30)
#endif

/* since screenshots are much bigger than cmds... */
#define T_REC_SHOTFILE_MAXSZ (1U << 30)

CARD32 thincRecordTimerShot(OsTimerPtr timer, CARD32 time, pointer arg);

/* msg specific functions */
int thinc_copyInitRecordStack(thincClientPtr cl);
void thinc_copyDestroyRecordStack(thincClientPtr cl);
int thinc_fillInitRecordStack(thincClientPtr cl);
void thinc_fillDestroyRecordStack(thincClientPtr cl);
int thinc_pixmapInitRecordStack(thincClientPtr cl);
void thinc_pixmapDestroyRecordStack(thincClientPtr cl);
int thinc_rawInitRecordStack(thincClientPtr cl);
void thinc_rawDestroyRecordStack(thincClientPtr cl);
int thinc_bitmapInitRecordStack(thincClientPtr cl);
void thinc_bitmapDestroyRecordStack(thincClientPtr cl);

/* cool macros. _out macros are meant to be used when writing data to the world
 * (e.g. they do htonl) */
/* copy thincRecCmd structs */
#define t_rec_copy_cmd_out(c1, c2) do {			\
	memcpy((c1).file, (c2).file, T_REC_FILEID_LEN);	\
	(c1).fp = htonl((c2).fp);				\
} while(0)

#define t_rec_copy_cmd(c1, c2) do {			\
	memcpy((c1).file, (c2).file, T_REC_FILEID_LEN);	\
	(c1).fp = (c2).fp;				\
} while(0)

/* fill a thincRecCmd struct */
#define t_rec_fill_cmd(c, fn, fpos) do {		\
	memcpy((c).file, fn, T_REC_FILEID_LEN);	\
	(c).fp = fpos;					\
} while(0)

#define t_rec_fill_cmd_out(c, fn, fpos) do {		\
	memcpy((c).file, fn, T_REC_FILEID_LEN);	\
	(c).fp = htonl(fpos);				\
} while(0)

/* fill a thincRecShot struct */
#define t_rec_fill_shot_out(s, fn, fpos) do {		\
	memcpy((s).file, fn, T_REC_FILEID_LEN);	\
	(s).fp = htonl(fpos);				\
} while(0)

void thinc_recordCurrCmd(thincScreenPtr pThinc, struct thincRecCmd *c_cmd);

#endif /* _THINC_RECORD_H_ */
