/* Copyright 2006, 2007 Ricardo Baratto */
/* Copyright 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
 *
 * Author:
 * Ricardo Baratto <ricardo@cs.columbia.edu>
 * Network Computing Lab - Columbia University
 */

/* 
 * this thing works by keeping a linked list of commands which are relevant to
 * the current contents of the screen. very similar to command queue struct for
 * now. The linked list is made of thincRecCommand structs 
 */

/* set tabstop=8 */

#if 0
#ifdef HAVE_XORG_CONFIG_H
#include <xorg-config.h>
#endif

#include "thinc.h"

#ifdef T_RECORD
static thincRecCmdPtr thincRecStateNew(thincRecordPtr record, BoxPtr box);
static thincRecCmdPtr thincRecStateNewReg(thincRecordPtr record, 
	RegionPtr reg);
static void thincRecStateDestroy(thincRecordPtr record, thincRecCmdPtr cmd);
static unsigned long thincRecGetID();

int thincRecordInitState(thincScreenPtr pThinc)
{
	thincRecordPtr record = t_recordPtr(pThinc);

	t_list_init(&record->state);

	return T_OK;
}

void thincRecordDestroyState(thincScreenPtr pThinc)
{
	thincRecordPtr record = t_recordPtr(pThinc);

	t_list_init(&record->state);
}

static thincRecCmdPtr thincRecStateNewReg(thincRecordPtr record, 
	RegionPtr reg)
{
	t_dbg_name(func)
	thincRecCmdPtr cmd;

	thincDbg(T_DBG_RECORD, ("%s: creating new command state\n", func));

	cmd = malloc(sizeof(*cmd));
	if(!cmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: No memory?\n", func));
		return NULL;
	}

	cmd->id = thincRecGetID();
	t_list_init(&cmd->list);
	memcpy(cmd->file, record->cmd_fn, T_REC_FILEID_LEN);
	cmd->fp = record->cmd_fp;

	REGION_INIT(record->pThinc->pScreen, &cmd->reg, NullBox, 0);
	REGION_COPY(record->pThinc->pScreen, &cmd->reg, reg);

	thincDbg(T_DBG_RECORD, ("%s: finished, new command %#lx\n", func,
		cmd->id));
	return cmd;
}

static thincRecCmdPtr thincRecStateNew(thincRecordPtr record, 
	BoxPtr box)
{
	thincRecCmdPtr cmd;
	RegionRec reg;

	REGION_INIT(record->pThinc->pScreen, &reg, box, 1);
	cmd = thincRecStateNewReg(record, &reg);
	REGION_UNINIT(record->pThinc->pScreen, &reg);

	return cmd;
}

static void thincRecStateDestroy(thincRecordPtr record, thincRecCmdPtr cmd)
{
	REGION_UNINIT(record->pThinc->pScreen, &cmd->reg);
	t_list_del_el(&cmd->list);
	free(cmd);
}

static unsigned long thincRecGetID()
{
	static unsigned long id = 0xFFFFFFFF;

	return --id;
}

#if 0
/********** state recorder *************/
int thincRecAddCommand(thincRecordPtr record, RegionPtr pReg)
{
#if 0
	t_dbg_name(func)
	thincRecCmdPtr cmd, next;
	RegionRec clip;
	BoxPtr currArea, exts= NULL;
	int modret, inreg, modop, chk_eq = T_FALSE;

	thincDbg(T_DBG_CMD, ("%s: Overwriting commands.\n", func));

	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, "overwrite region:");

	if(REGION_NUM_RECTS(pReg) == 1) {
		chk_eq = T_TRUE;
		exts = REGION_EXTENTS(pThinc->pScreen, pReg);
	}

	t_list_for_each_entry(cmd, next, thincCmdRec, list, list) {
		thincDbg(T_DBG_CMD, ("%s: id %#lx, type %d\n",
			func, cmd->id, cmd->type));

#if 0
		if(cmd->flags & T_CMD_NO_OVER) {
			/* "can't touch this" */
			thincDbg(T_DBG_CMD, ("%s: found untouchable "
				"command. skipping\n", func));
			continue;
		}
#endif

		currArea = &cmd->area;

		print_box(T_DBG_CMD, currArea, "command's area:");

		if(chk_eq && box_equal(exts, currArea)) {
			thincDbg(T_DBG_CMD, ("%s: command's area is same "
				"as new one. you're outa hier!\n", func));
			/* XXX: This is kind of sketchy, but needed. In
			 * theory, we should just delete the command from
			 * the actual list we're passed. but:
			 * 1. we don't know if that list is cmd->list or
			 * cmd->sched
			 * 2. sometimes we need to delete it from both.
			 */
			thincDeleteCmdAll(pThinc, cmd);
			t_list_len_dec(list);
			continue;
		}
		/*
		 * this checks if currArea is inside, outside or miti-miti
		 * of pReg.
		 * then:
		 * -> if currArea is inside pReg, then the command is
		 *  completely clipped out and we should just delete it.
		 * -> if currArea is outside pReg, then we don't have
		 *  to do anything with this dud so just go to next
		 *  command.
		 * -> if there's some overlap but no clear winner, then
		 *  need to clip the command with currReg - pReg
		 */
		inreg = RECT_IN_REGION(pThinc->pScreen, pReg, currArea);
		if(inreg == rgnOUT) {
			/* sweet, next */
			thincDbg(T_DBG_CMD, ("%s: current command is not "
				"overwritten by region, ignoring\n", func));
			continue;
		} else if(inreg == rgnIN) {
			thincDbg(T_DBG_CMD, ("%s: current command "
				"completely obscured by new command. "
				"Deleting from list\n", func));
			/* XXX: This is kind of sketchy, but needed. In
			 * theory, we should just delete the command from
			 * the actual list we're passed. but:
			 * 1. we don't know if that list is cmd->list or
			 * cmd->sched
			 * 2. sometimes we need to delete it from both.
			 */
			thincDeleteCmdAll(pThinc, cmd);
			t_list_len_dec(list);
			continue;
		}

		/* inreg == rgnPART */

		/* there's overlap. Need to modify this command */
		thincDbg(T_DBG_CMD, ("%s: Command conficts with region. "
			"Modifying\n", func));

		REGION_INIT(pThinc->pScreen, &clip, currArea, 1);
		REGION_SUBTRACT(pThinc->pScreen, &clip, &clip, pReg);
		print_reg(T_DBG_CMD, pThinc->pScreen, &clip,
			"clipping command with: ");

		modop = T_CMD_OVER;
		modret = thincModifyCmdInList(pThinc, cmd, modop, &clip);
		REGION_UNINIT(pThinc->pScreen, &clip);
		/* don't care about the other return values since we
		 * don't do anything else. Note that if modret is 0,
		 * curr is no longer valid, so be careful if you add
		 * code before the end of the loop */
		if(modret == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't modify "
				"command in queue. Destroying it\n", func));
			/* XXX: This is kind of sketchy, but needed. In
			 * theory, we should just delete the command from
			 * the actual list we're passed. but:
			 * 1. we don't know if that list is cmd->list or
			 * cmd->sched
			 * 2. sometimes we need to delete it from both
			 * (e.g. when overwriting on the client buffer,
			 * and the command is already scheduled)
			 */
			thincDeleteCmdAll(pThinc, cmd);
			t_list_len_dec(list);
			continue;
		} else if(modret == T_CLIPPED) {
			/* Modify() doesn't have enough information to
			 * do this so here we are */
			t_list_len_dec(list);
		}
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
#endif
	return T_OK;
}
#endif

#endif
#endif
